from functools import partial

import numpy as np
from scipy.spatial.distance import pdist, squareform

from ..tools import compute_dist, perm_test
from ._utils import _CheckInputs
from .base import ConditionalIndependenceTest, ConditionalIndependenceTestOutput


class ConditionalDcorr(ConditionalIndependenceTest):
    r"""
    Conditional Distance Covariance/Correlation (CDcov/CDcorr) test statistic and p-value.

    CDcorr is a measure of dependence between two paired random matrices
    given a third random matrix of not necessarily equal dimensions :footcite:p:`wang2015conditional`.
    The coefficient is 0 if and only if the matrices are independent given
    third matrix.

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already distance
        matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguements to send to your custom
        function.
    use_cov : bool,
        If `True`, then the statistic will compute the covariance rather than the
        correlation.
    bandwith : str, scalar, 1d-array
        The method used to calculate the bandwidth used for kernel density estimate of
        the conditional matrix. This can be ‘scott’, ‘silverman’, a scalar constant or a
        1d-array with length ``r`` which is the dimensions of the conditional matrix.
        If None (default), ‘scott’ is used.
    **kwargs
        Arbitrary keyword arguments for ``compute_distance``.

    References
    ----------
    .. footbibliography::
    """

    def __init__(
        self, compute_distance="euclidean", use_cov=True, bandwidth=None, **kwargs
    ):
        self.use_cov = use_cov
        self.compute_distance = compute_distance
        self.bandwidth = bandwidth

        # Check bandwidth input
        if bandwidth is not None:
            if not isinstance(bandwidth, (int, float, np.ndarray, str)):
                raise ValueError(
                    "`bandwidth` should be 'scott', 'silverman', a scalar or 1d-array."
                )
            if isinstance(bandwidth, str):
                if bandwidth.lower() not in ["scott", "silverman"]:
                    raise ValueError(
                        f"`bandwidth` must be either 'scott' or 'silverman' not '{bandwidth}'"
                    )
                self.bandwidth = bandwidth.lower()

        # set is_distance to true if compute_distance is None
        self.is_distance = False
        if not compute_distance:
            self.is_distance = True

        ConditionalIndependenceTest.__init__(self, **kwargs)

    def __repr__(self):
        return "ConditionalDcorr"

    def statistic(self, x, y, z):
        r"""
        Helper function that calculates the CDcov/CDcorr test statistic.

        Parameters
        ----------
        x,y,z : ndarray of float
            Input data matrices. ``x``, ``y`` and ``z`` must have the same number
            of samples. That is, the shapes must be ``(n, p)``, ``(n, q)`` and
            ``(n, r)`` where `n` is the number of samples and `p`, `q`, and `r`
            are the number of dimensions. Alternatively, ``x`` and ``y`` can be
            distance matrices and ``z`` can be a similarity matrix where the
            shapes must be ``(n, n)``.

        Returns
        -------
        stat : float
            The computed CDcov/CDcorr statistic.
        """
        check_input = _CheckInputs(x, y, z, ignore_z_var=True)
        x, y, z = check_input()

        if not self.is_distance:
            distx, disty = compute_dist(
                x,
                y,
                metric=self.compute_distance,
            )
            distz = self._compute_kde(z)
        else:
            distx = x
            disty = y
            distz = z

        if self.use_cov:
            stat = _cdcov(distx, disty, distz).mean()
        else:
            stat = _cdcorr(distx, disty, distz).mean()

        self.stat = stat
        return stat

    def test(
        self,
        x,
        y,
        z,
        reps=1000,
        workers=1,
        random_state=None,
    ):
        r"""
        Calculates the CDcov/CDcorr test statistic and p-value.

        Parameters
        ----------
        x,y,z : ndarray of float
            Input data matrices. ``x``, ``y`` and ``z`` must have the same number
            of samples. That is, the shapes must be ``(n, p)``, ``(n, q)`` and
            ``(n, r)`` where `n` is the number of samples and `p`, `q`, and `r`
            are the number of dimensions. Alternatively, ``x`` and ``y`` can be
            distance matrices and ``z`` can be a similarity matrix where the
            shapes must be ``(n, n)``.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        random_state : int, default: None
            The random_state for permutation testing to be fixed for
            reproducibility.

        Returns
        -------
        stat : float
            The computed CDcov/CDcorr statistic.
        pvalue : float
            The computed CDcov/CDcorr p-value.
        """
        check_input = _CheckInputs(x, y, z, reps=reps, ignore_z_var=True)
        x, y, z = check_input()

        if not self.is_distance:
            x, y = compute_dist(x, y, metric=self.compute_distance, **self.kwargs)
            z = self._compute_kde(z)

            self.is_distance = True

        stat, pvalue, null_dist = perm_test(
            self.statistic,
            x=x,
            y=y,
            z=z,
            reps=reps,
            workers=workers,
            is_distsim=self.is_distance,
            random_state=random_state,
            permuter=partial(self._permuter, probs=z),
        )
        self.stat = stat
        self.pvalue = pvalue
        self.null_dist = null_dist

        return ConditionalIndependenceTestOutput(stat, pvalue)

    def _compute_kde(self, data):
        n, d = data.shape

        if isinstance(self.bandwidth, (int, float)):
            bandwidth_ = np.repeat(self.bandwidth, d)
        elif isinstance(self.bandwidth, np.ndarray):
            if not d == self.bandwidth.size:
                raise ValueError(f"`bandwidth` must be of length {d}.")
            bandwidth_ = self.bandwidth
        elif self.bandwidth is None or self.bandwidth == "scott":
            factor = np.power(n, (-1.0 / (d + 4)))
            stds = np.std(data, axis=0, ddof=1)
            bandwidth_ = factor * stds
        elif self.bandwidth == "silverman":
            factor = (n * (d + 2) / 4.0) ** (-1.0 / (d + 4))
            stds = np.std(data, axis=0, ddof=1)
            bandwidth_ = factor * stds

        # Compute constants
        denom = np.power(2 * np.pi, d / 2) * np.power(bandwidth_.prod(), 0.5)

        # Compute weighted kernel
        sim_mat = pdist(
            data, "sqeuclidean", w=1 / bandwidth_**2
        )  # Section 4.2 equation
        sim_mat = squareform(-0.5 * sim_mat)
        np.exp(sim_mat, sim_mat)
        sim_mat /= denom

        return sim_mat

    def _permuter(self, probs, rng=None, axis=1):
        """
        Weighted sampling with replacement
        """
        if rng is None:
            rng = np.random.default_rng()

        n = probs.shape[1 - axis]
        sums = probs.sum(axis=1, keepdims=True)
        idx = ((probs / sums).cumsum(axis=1) > rng.random(n)[:, None]).argmax(axis=1)

        return idx


def _weighted_center_distmat(distx, weights): # pragma: no cover
    """Centers the distance matrices with weights"""

    n = distx.shape[0]

    scl = np.sum(weights)
    row_sum = np.sum(np.multiply(distx, weights), axis=1) / scl
    total_sum = weights @ row_sum / scl

    cent_distx = distx - row_sum.reshape(-1, n).T - row_sum.reshape(-1, n) + total_sum

    return cent_distx


def _cdcov(distx, disty, distz): # pragma: no cover
    """Calculate the CDcov test statistic"""
    n = distx.shape[0]

    cdcov = np.zeros(n)

    for i in range(n):
        r = distz[[i]]
        cdx = _weighted_center_distmat(distx, distz[i])
        cdy = _weighted_center_distmat(disty, distz[i])
        cdcov[i] = (cdx * cdy * r * r.T).sum() / r.sum() ** 2

    cdcov *= 12 * np.power(distz.mean(axis=0), 4)

    return cdcov


def _cdcorr(distx, disty, distz): # pragma: no cover
    """Calculate the CDcorr test statistic"""
    varx = _cdcov(distx, distx, distz)
    vary = _cdcov(disty, disty, distz)
    covar = _cdcov(distx, disty, distz)

    # stat is 0 with negative variances (would make denominator undefined)
    denom = varx * vary
    non_positives = denom <= 0
    if np.any(non_positives):
        denom[non_positives] = 0
    stat = np.divide(
        covar, np.sqrt(denom), out=np.zeros_like(covar), where=np.invert(non_positives)
    )

    return stat



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from hyppo.tools import rot_ksamp
from hyppo.ksample import SmoothCFTest


class TestSmoothCF:
    @pytest.mark.parametrize(
        "n, obs_stat, obs_pvalue",
        [
            (2000, 1317.0159, 8.134e-277),
            (1000, 546.279, 5.608e-111),
        ],
    )
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = rot_ksamp("linear", n, 1, k=2)
        stat, pvalue = SmoothCFTest().test(x, y)

        assert_almost_equal(stat, obs_stat, decimal=1)
        assert_almost_equal(pvalue, obs_pvalue, decimal=100)

    def test_null(self):
        np.random.seed(120)
        x = np.random.randn(500, 10)
        y = np.random.randn(500, 10)
        _, pval = SmoothCFTest().test(x, y)
        assert pval > 0.05

    def test_alternative(self):
        np.random.seed(120)
        x = np.random.randn(500, 10)
        x[:, 1] *= 3
        y = np.random.randn(500, 10)
        _, pval = SmoothCFTest().test(x, y)
        assert pval < 0.05

    @pytest.mark.parametrize(
        "obs_stat, obs_pval",
        [
            (0.2707, 0.8734),
        ],
    )
    def test_one_frequency(self, obs_stat, obs_pval):
        np.random.seed(123456789)
        x = np.random.randn(500, 10)
        y = np.random.randn(500, 10)
        stat, pvalue = SmoothCFTest(num_randfreq=1).test(x, y, random_state=1234)
        assert_almost_equal(stat, obs_stat, decimal=2)
        assert_almost_equal(pvalue, obs_pval, decimal=2)



import numpy as np
from numba import jit

from ..tools import check_perm_blocks_dim, chi2_approx, compute_dist
from ._utils import _CheckInputs
from .base import IndependenceTest, IndependenceTestOutput


class Dcorr(IndependenceTest):
    r"""
    Distance Correlation (Dcorr) test statistic and p-value.

    Dcorr is a measure of dependence between two paired random matrices of
    not necessarily equal dimensions. The coefficient is 0 if and only if the
    matrices are independent. It is an example of an energy distance.

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already distance
        matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguements to send to your custom
        function.
    bias : bool, default: False
        Whether or not to use the biased or unbiased test statistics.
    **kwargs
        Arbitrary keyword arguments for ``compute_distance``.

    Notes
    -----
    The statistic can be derived as follows:

    Let :math:`x` and :math:`y` be :math:`(n, p)` samples of random variables
    :math:`X` and :math:`Y`. Let :math:`D^x` be the :math:`n \times n`
    distance matrix of :math:`x` and :math:`D^y` be the :math:`n \times n` be
    the distance matrix of :math:`y`. The distance covariance is,

    .. math::

        \mathrm{Dcov}^b_n (x, y) = \frac{1}{n^2} \mathrm{tr} (H D^x H H D^y H)

    where :math:`\mathrm{tr} (\cdot)` is the trace operator and :math:`H` is
    defined as :math:`H = I - (1/n) J` where :math:`I` is the identity matrix
    and :math:`J` is a matrix of ones. The normalized version of this
    covariance is distance correlation
    :footcite:p:`szekelyMeasuringTestingDependence2007` and is

    .. math::

        \mathrm{Dcorr}^b_n (x, y) = \frac{\mathrm{Dcov}^b_n (x, y)}
                                       {\sqrt{\mathrm{Dcov}^b_n (x, x)
                                              \mathrm{Dcov}^b_n (y, y)}}

    This is a biased test statistic. An unbiased alternative also exists, and is
    defined using the following: Consider the
    centering process where :math:`\mathbb{1}(\cdot)` is the indicator
    function:

    .. math::

        C^x_{ij} = \left[ D^x_{ij} - \frac{1}{n-2} \sum_{t=1}^n D^x_{it}
            - \frac{1}{n-2} \sum_{s=1}^n D^x_{sj}
            + \frac{1}{(n-1) (n-2)} \sum_{s,t=1}^n D^x_{st} \right]
            \mathbb{1}_{i \neq j}

    and similarly for :math:`C^y`. Then, this unbiased Dcorr is,

    .. math::

        \mathrm{Dcov}_n (x, y) = \frac{1}{n (n-3)} \mathrm{tr} (C^x C^y)

    The normalized version of this covariance
    :footcite:p:`szekelyPartialDistanceCorrelation2014a` is

    .. math::

        \mathrm{Dcorr}_n (x, y) = \frac{\mathrm{Dcov}_n (x, y)}
                                        {\sqrt{\mathrm{Dcov}_n (x, x)
                                               \mathrm{Dcov}_n (y, y)}}

    The p-value returned is calculated using a permutation test using
    :meth:`hyppo.tools.perm_test`. The fast version of the test uses
    :meth:`hyppo.tools.chi2_approx`.

    When the data is 1 dimension and the distance metric is Euclidean,
    and even faster version of the algorithm is run (computational
    complexity is :math:`\mathcal{O}(n \log n)`)
    :footcite:p:`chaudhuriFastAlgorithmComputing2019`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, compute_distance="euclidean", bias=False, **kwargs):
        # set is_distance to true if compute_distance is None
        self.is_distance = False
        if not compute_distance:
            self.is_distance = True
        self.bias = bias
        self.is_fast = False
        IndependenceTest.__init__(self, compute_distance=compute_distance, **kwargs)

    def statistic(self, x, y):
        r"""
        Helper function that calculates the Dcorr test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.

        Returns
        -------
        stat : float
            The computed Dcorr statistic.
        """
        distx = x
        disty = y

        if not (self.is_distance or self.is_fast):
            distx, disty = compute_dist(
                x, y, metric=self.compute_distance, **self.kwargs
            )

        stat = _dcorr(distx, disty, bias=self.bias, is_fast=self.is_fast)
        self.stat = stat

        return stat

    def test(
        self, x, y, reps=1000, workers=1, auto=True, perm_blocks=None, random_state=None
    ):
        r"""
        Calculates the Dcorr test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        auto : bool, default: True
            Automatically uses fast approximation when `n` and size of array
            is greater than 20. If ``True``, and sample size is greater than 20, then
            :class:`hyppo.tools.chi2_approx` will be run. Parameters ``reps`` and
            ``workers`` are
            irrelevant in this case. Otherwise, :class:`hyppo.tools.perm_test` will be
            run.
            If ``x`` and ``y`` have `p` equal to 1 and ``compute_distance`` set to
            ``'euclidean'``, then and :math:`\mathcal{O}(n \log n)` version is run.
        perm_blocks : None or ndarray, default: None
            Defines blocks of exchangeable samples during the permutation test.
            If None, all samples can be permuted with one another. Requires `n`
            rows. At each column, samples with matching column value are
            recursively partitioned into blocks of samples. Within each final
            block, samples are exchangeable. Blocks of samples from the same
            partition are also exchangeable between one another. If a column
            value is negative, that block is fixed and cannot be exchanged.

        Returns
        -------
        stat : float
            The computed Dcorr statistic.
        pvalue : float
            The computed Dcorr p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.independence import Dcorr
        >>> x = np.arange(25)
        >>> y = x
        >>> stat, pvalue = Dcorr().test(x, y)
        >>> '%.1f, %.2f' % (stat, pvalue)
        '1.0, 0.00'

        In addition, the inputs can be distance matrices. In this case,
        the ``compute_distance`` parameter must be set
        to ``None``.

        >>> import numpy as np
        >>> from hyppo.independence import Dcorr
        >>> x = np.ones((10, 10)) - np.identity(10)
        >>> y = 2 * x
        >>> dcorr = Dcorr(compute_distance=None)
        >>> stat, pvalue = dcorr.test(x, y)
        >>> '%.1f, %.2f' % (stat, pvalue)
        '0.0, 1.00'
        """
        check_input = _CheckInputs(
            x,
            y,
            reps=reps,
        )
        x, y = check_input()
        if perm_blocks is not None:
            check_perm_blocks_dim(perm_blocks, y)

        if (
            auto
            and x.shape[1] == 1
            and y.shape[1] == 1
            and self.compute_distance == "euclidean"
        ):
            self.is_fast = True

        if auto and x.shape[0] > 20 and perm_blocks is None and not self.bias:
            stat, pvalue = chi2_approx(self.statistic, x, y)
            self.stat = stat
            self.pvalue = pvalue
            self.null_dist = None
        else:
            if not self.is_fast:
                x, y = compute_dist(x, y, metric=self.compute_distance, **self.kwargs)
                self.is_distance = True
            stat, pvalue = super(Dcorr, self).test(
                x,
                y,
                reps,
                workers,
                perm_blocks=perm_blocks,
                is_distsim=self.is_distance,
                random_state=random_state,
            )

        return IndependenceTestOutput(stat, pvalue)


@jit(nopython=True, cache=True)
def _center_distmat(distx, bias):  # pragma: no cover
    """Centers the distance matrices"""
    n = distx.shape[0]
    if bias:
        # use sum instead of mean because of numba restrictions
        exp_distx = (
            np.repeat(distx.sum(axis=0) / n, n).reshape(-1, n).T
            + np.repeat(distx.sum(axis=1) / n, n).reshape(-1, n)
            - (distx.sum() / (n * n))
        )
    else:
        exp_distx = (
            np.repeat((distx.sum(axis=0) / (n - 2)), n).reshape(-1, n).T
            + np.repeat((distx.sum(axis=1) / (n - 2)), n).reshape(-1, n)
            - distx.sum() / ((n - 1) * (n - 2))
        )
    cent_distx = distx - exp_distx
    if not bias:
        np.fill_diagonal(cent_distx, 0)
    return cent_distx


@jit(nopython=True, cache=True)
def _cpu_cumsum(data):  # pragma: no cover
    """Create cumulative sum since numba doesn't sum over axes."""
    cumsum = data.copy()
    for i in range(1, data.shape[0]):
        cumsum[i, :] = data[i, :] + cumsum[i - 1, :]
    return cumsum


@jit(nopython=True, cache=True)
def _fast_1d_dcov(x, y, bias=False):  # pragma: no cover
    """
    Calculate the Dcorr test statistic. Note that though Dcov is calculated
    and stored in covar, but not called due to a slower implementation.

    See: https://www.sciencedirect.com/science/article/abs/pii/S0167947319300313
    """
    n = x.shape[0]

    # sort inputs
    x_orig = x.ravel()
    x = np.sort(x_orig)
    y = y[np.argsort(x_orig)]
    x = x.reshape(-1, 1)  # for numba

    # cumulative sum
    si = _cpu_cumsum(x)
    ax = (np.arange(-(n - 2), n + 1, 2) * x.ravel()).reshape(-1, 1) + (si[-1] - 2 * si)

    v = np.hstack((x, y, x * y))
    nw = v.shape[1]

    idx = np.vstack((np.arange(n), np.zeros(n))).astype(np.int64).T
    iv1 = np.zeros((n, 1))
    iv2 = np.zeros((n, 1))
    iv3 = np.zeros((n, 1))
    iv4 = np.zeros((n, 1))

    i = 1
    r = 0
    s = 1
    while i < n:
        gap = 2 * i
        k = 0
        idx_r = idx[:, r]
        csumv = np.vstack((np.zeros((1, nw)), _cpu_cumsum(v[idx_r, :])))

        for j in range(1, n + 1, gap):
            st1 = j - 1
            e1 = min(st1 + i - 1, n - 1)
            st2 = j + i - 1
            e2 = min(st2 + i - 1, n - 1)

            while (st1 <= e1) and (st2 <= e2):
                idx1 = idx_r[st1]
                idx2 = idx_r[st2]

                if y[idx1] >= y[idx2]:
                    idx[k, s] = idx1
                    st1 += 1
                else:
                    idx[k, s] = idx2
                    st2 += 1
                    iv1[idx2] += e1 - st1 + 1
                    iv2[idx2] += csumv[e1 + 1, 0] - csumv[st1, 0]
                    iv3[idx2] += csumv[e1 + 1, 1] - csumv[st1, 1]
                    iv4[idx2] += csumv[e1 + 1, 2] - csumv[st1, 2]
                k += 1

            if st1 <= e1:
                kf = k + e1 - st1 + 1
                idx[k:kf, s] = idx_r[st1 : e1 + 1]
                k = kf
            elif st2 <= e2:
                kf = k + e2 - st2 + 1
                idx[k:kf, s] = idx_r[st2 : e2 + 1]
                k = kf

        i = gap
        r = 1 - r
        s = 1 - s

    covterm = np.sum(n * (x - np.mean(x)).T @ (y - np.mean(y)))
    c1 = np.sum(iv1.T @ v[:, 2].copy())
    c2 = np.sum(iv4)
    c3 = np.sum(iv2.T @ y)
    c4 = np.sum(iv3.T @ x)
    d = 4 * ((c1 + c2) - (c3 + c4)) - 2 * covterm

    y_sorted = y[idx[n::-1, r], :]
    si = _cpu_cumsum(y_sorted)
    by = np.zeros((n, 1))
    by[idx[::-1, r]] = (np.arange(-(n - 2), n + 1, 2) * y_sorted.ravel()).reshape(
        -1, 1
    ) + (si[-1] - 2 * si)

    if bias:
        denom = [n**2, n**3, n**4]
    else:
        denom = [n * (n - 3), n * (n - 3) * (n - 2), n * (n - 3) * (n - 2) * (n - 1)]

    stat = np.sum(
        (d / denom[0])
        + (np.sum(ax) * np.sum(by) / denom[2])
        - (2 * (ax.T @ by) / denom[1])
    )

    return stat


@jit(nopython=True, cache=True)
def _dcov(distx, disty, bias=False, only_dcov=True):  # pragma: no cover
    """Calculate the Dcov test statistic"""
    if only_dcov:
        # center distance matrices
        distx = _center_distmat(distx, bias)
        disty = _center_distmat(disty, bias)

    stat = np.sum(distx * disty)

    if only_dcov:
        N = distx.shape[0]
        if bias:
            # only biased Dcov is square-rooted
            # https://search.r-project.org/CRAN/refmans/energy/html/dcovu.html
            stat = np.sqrt(1 / (N**2) * stat)
        else:
            stat = 1 / (N * (N - 3)) * stat

    return stat


@jit(nopython=True, cache=True)
def _dcorr(distx, disty, bias=False, is_fast=False):  # pragma: no cover
    """
    Calculate the Dcorr test statistic.
    """
    if is_fast:
        # calculate covariances and variances
        covar = _fast_1d_dcov(distx, disty, bias=bias)
        varx = _fast_1d_dcov(distx, distx, bias=bias)
        vary = _fast_1d_dcov(disty, disty, bias=bias)
    else:
        # center distance matrices
        distx = _center_distmat(distx, bias)
        disty = _center_distmat(disty, bias)

        # calculate covariances and variances
        covar = _dcov(distx, disty, bias=bias, only_dcov=False)
        varx = _dcov(distx, distx, bias=bias, only_dcov=False)
        vary = _dcov(disty, disty, bias=bias, only_dcov=False)

    # stat is 0 with negative variances (would make denominator undefined)
    if varx <= 0 or vary <= 0 or covar <= 0:
        stat = 0

    # calculate generalized test statistic
    else:
        stat = covar / np.sqrt(np.real(varx * vary))
        # only biased Dcov is square-rooted
        # https://search.r-project.org/CRAN/refmans/energy/html/dcovu.html
        if bias:
            stat = np.sqrt(stat)

    return stat



import warnings

import numpy as np
from joblib import Parallel, delayed
from scipy.stats.distributions import chi2
from sklearn.metrics import pairwise_distances
from sklearn.metrics.pairwise import pairwise_kernels
from sklearn.utils import check_random_state


# Explicitly copying private function from scipy 1.7.3
# Modified to only use nan_policy 'raise'
# REF: https://github.com/scipy/scipy/blob/59e6539cf80dc04b16b0f0ab52343381f0a7a2fa/scipy/stats/stats.py#L79
def contains_nan(a):
    nan_policy = "raise"
    try:
        # Calling np.sum to avoid creating a huge array into memory
        # e.g. np.isnan(a).any()
        with np.errstate(invalid="ignore"):
            contains_nan_var = np.isnan(np.sum(a))
    except TypeError:
        # This can happen when attempting to sum things which are not
        # numbers (e.g. as in the function `mode`). Try an alternative method:
        try:
            contains_nan_var = np.nan in set(a.ravel())
        except TypeError:
            # Don't know what to do. Fall back to omitting nan values and
            # issue a warning.
            contains_nan_var = False
            nan_policy = "omit"
            warnings.warn(
                "The input array could not be properly "
                "checked for nan values. nan values "
                "will be ignored.",
                RuntimeWarning,
            )

    if contains_nan_var:
        raise ValueError("The input contains nan values")

    return contains_nan_var, nan_policy


def check_ndarray_xy(x, y):
    """Check if x or y is an ndarray of float"""
    if not isinstance(x, np.ndarray) or not isinstance(y, np.ndarray):
        raise TypeError("x and y must be ndarrays")


def check_ndarray_xyz(x, y, z):
    """Check if x or y is an ndarray of float"""
    if (
        not isinstance(x, np.ndarray)
        or not isinstance(y, np.ndarray)
        or not isinstance(z, np.ndarray)
    ):
        raise TypeError("x, y, and z must be ndarrays")


def convert_xy_float64(x, y):
    """Convert x or y to np.float64 (if not already done)"""
    # convert x and y to floats
    x = np.asarray(x).astype(np.float64)
    y = np.asarray(y).astype(np.float64)

    return x, y


def convert_xyz_float64(x, y, z):
    """Convert x or y or z to np.float64 (if not already done)"""
    # convert x and y to floats
    x = np.asarray(x).astype(np.float64)
    y = np.asarray(y).astype(np.float64)
    z = np.asarray(z).astype(np.float64)

    return x, y, z


def check_reps(reps):
    """Check if reps is valid"""
    # check if reps is an integer > than 0
    if not isinstance(reps, int) or reps < 0:
        raise ValueError("Number of reps must be an integer greater than 0.")

    # check if reps is under 1000 (recommended)
    elif reps < 1000:
        msg = (
            "The number of replications is low (under 1000), and p-value "
            "calculations may be unreliable. Use the p-value result, with "
            "caution!"
        )
        warnings.warn(msg, RuntimeWarning)


def _check_distmat(x, y, z=None):
    """Check if x, y, and z are distance matrices."""
    if z is None:
        data = (x, y)
        labs = np.array(["x", "y"])
    else:
        data = (x, y, z)
        labs = np.array(["x", "y", "z"])

    check_sym = np.array([not np.allclose(arr, arr.T) for arr in data])
    check_diag = np.array([not np.all((arr.diagonal() == 0)) for arr in data])

    if np.any(check_sym) or np.any(check_diag):
        inputs = "x and y" if z is None else "x, y and z"
        if np.any(check_sym):
            names = ", ".join(labs[check_sym])
            verb = "is" if check_sym.sum() == 1 else "are"
            sym_msg = f"{names} {verb} not symmetric."
        else:
            sym_msg = ""
        if np.any(check_diag):
            names = ", ".join(labs[check_diag])
            verb = "does not" if check_diag.sum() == 1 else "do not"
            diag_msg = f"{names} {verb} have zeros along the diagonal."
        else:
            diag_msg = ""

        msg = f"{inputs} must be distance matrices. {sym_msg} {diag_msg}"
        raise ValueError(msg)


def _check_kernmat(x, y):
    """Check if x and y are similarity matrices."""
    if (
        not np.allclose(x, x.T)
        or not np.allclose(y, y.T)
        or not np.all((x.diagonal() == 1))
        or not np.all((y.diagonal() == 1))
    ):
        raise ValueError(
            "x and y must be kernel similarity matrices, "
            "{is_sym} symmetric and {one_diag} "
            "ones along the diagonal".format(
                is_sym="x is not"
                if not np.array_equal(x, x.T)
                else "y is not"
                if not np.array_equal(y, y.T)
                else "both are",
                one_diag="x doesn't have"
                if not np.all((x.diagonal() == 1))
                else "y doesn't have"
                if not np.all((y.diagonal() == 1))
                else "both have",
            )
        )


def compute_kern(x, y, metric="gaussian", workers=1, **kwargs):
    """
    Kernel similarity matrices for the inputs.

    Parameters
    ----------
    x,y : ndarray of float
        Input data matrices. ``x`` and ``y`` must have the same number of
        samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
        `n` is the number of samples and `p` and `q` are the number of
        dimensions. Alternatively, ``x`` and ``y`` can be kernel similarity matrices,
        where the shapes must both be ``(n, n)``.
    metric : str, callable, or None, default: "gaussian"
        A function that computes the kernel similarity among the samples within each
        data matrix.
        Valid strings for ``metric`` are, as defined in
        :func:`sklearn.metrics.pairwise.pairwise_kernels`,

            [``"additive_chi2"``, ``"chi2"``, ``"linear"``, ``"poly"``,
            ``"polynomial"``, ``"rbf"``,
            ``"laplacian"``, ``"sigmoid"``, ``"cosine"``]

        Note ``"rbf"`` and ``"gaussian"`` are the same metric.
        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already similarity
        matrices. To call a custom function, either create the similarity matrix
        before-hand or create a function of the form :func:`metric(x, **kwargs)`
        where ``x`` is the data matrix for which pairwise kernel similarity matrices are
        calculated and kwargs are extra arguements to send to your custom function.
    workers : int, default: 1
        The number of cores to parallelize the p-value computation over.
        Supply ``-1`` to use all cores available to the Process.
    **kwargs
        Arbitrary keyword arguments provided to
        :func:`sklearn.metrics.pairwise.pairwise_kernels`
        or a custom kernel function.

    Returns
    -------
    simx, simy : ndarray of float
        Similarity matrices based on the metric provided by the user.
    """
    if not metric:
        metric = "precomputed"
    if metric in ["gaussian", "rbf"]:
        if "gamma" not in kwargs:
            l2 = pairwise_distances(x, metric="l2", n_jobs=workers)
            n = l2.shape[0]
            # compute median of off diagonal elements
            med = np.median(
                np.lib.stride_tricks.as_strided(
                    l2, (n - 1, n + 1), (l2.itemsize * (n + 1), l2.itemsize)
                )[:, 1:]
            )
            # prevents division by zero when used on label vectors
            med = med if med else 1
            kwargs["gamma"] = 1.0 / (2 * (med**2))
        metric = "rbf"
    if callable(metric):
        simx = metric(x, **kwargs)
        simy = metric(y, **kwargs)
        _check_kernmat(
            simx, simy
        )  # verify whether matrix is correct, built into sklearn func
    else:
        simx = pairwise_kernels(x, metric=metric, n_jobs=workers, **kwargs)
        simy = pairwise_kernels(y, metric=metric, n_jobs=workers, **kwargs)
    return simx, simy


def _multi_check_kernmat(*args):
    """Check if every input is a similarity matrix."""
    for x in args:
        if not np.allclose(x, x.T) or not np.all((x.diagonal() == 1)):
            raise ValueError(
                "x must be a kernel similarity matrix, "
                "{is_sym} symmetric and {one_diag} "
                "ones along the diagonal".format(
                    is_sym="is not" if not np.array_equal(x, x.T) else "is",
                    one_diag="doesn't have"
                    if not np.all((x.diagonal() == 1))
                    else "has",
                )
            )


def multi_compute_kern(*args, metric="gaussian", workers=1, **kwargs):
    """
    Kernel similarity matrices for the input matrices.

    Parameters
    ----------
    *args: ndarray of float
        Variable length input data matrices. All inputs must have the same
        number of samples. That is, the shapes must be ``(n, p)``, ``(n, q)``,
        etc., where `n` is the number of samples and `p` and `q` are the
        number of dimensions.
    metric: str, callable, or None, default="gaussian"
        A function that computes the kernel similarity among the samples within each
        data matrix.
        Valid strings for ``metric`` are, as defined in
        :func:`sklearn.metrics.pairwise.pairwise_kernels`,

            [``"additive_chi2"``, ``"chi2"``, ``"linear"``, ``"poly"``,
            ``"polynomial"``, ``"rbf"``,
            ``"laplacian"``, ``"sigmoid"``, ``"cosine"``]

        Note ``"rbf"`` and ``"gaussian"`` are the same metric.
        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already similarity
        matrices. To call a custom function, either create the similarity matrix
        before-hand or create a function of the form :func:`metric(x, **kwargs)`
        where ``x`` is the data matrix for which pairwise kernel similarity matrices are
        calculated and kwargs are extra arguements to send to your custom function.
    workers: int, default=1
        The number of cores to parallelize the p-value computation over.
        Supply ``-1`` to use all cores available to the Process.
    **kwargs
        Arbitrary keyword arguments provided to
        :func:`sklearn.metrics.pairwise.pairwise_kernels`
        or a custom kernel function.

    Returns
    -------
    sim_matrices: ndarray of float
        Similarity matrices based on the metric provided by the user.
        Must be same shape as ''args''.
    """
    if not metric:
        metric = "precomputed"
    if metric in ["gaussian", "rbf"]:
        if "gamma" not in kwargs:
            l2 = pairwise_distances(args[0], metric="l2", n_jobs=workers)
            n = l2.shape[0]
            # compute median of off diagonal elements
            med = np.median(
                np.lib.stride_tricks.as_strided(
                    l2, (n - 1, n + 1), (l2.itemsize * (n + 1), l2.itemsize)
                )[:, 1:]
            )
            # prevents division by zero when used on label vectors
            med = med if med else 1
            kwargs["gamma"] = 1.0 / (2 * (med**2))
        metric = "rbf"
    if callable(metric):
        sim_mats = []
        for mat in args:
            sim_mat = metric(mat, **kwargs)
            sim_mats.append(sim_mat)
        sim_matrices = tuple(sim_mats)
        _multi_check_kernmat(*sim_matrices)
    else:
        sim_mats = []
        for mat in args:
            sim_mat = pairwise_kernels(mat, metric=metric, n_jobs=workers, **kwargs)
            sim_mats.append(sim_mat)
        sim_matrices = tuple(sim_mats)
    return sim_matrices


def compute_dist(x, y, z=None, metric="euclidean", workers=1, **kwargs):
    """
    Distance matrices for the inputs.

    Parameters
    ----------
    x,y,z : ndarray of float
        Input data matrices. ``x``, ``y`` and ``z`` must have the same number
        of samples. That is, the shapes must be ``(n, p)``, ``(n, q)`` and
        ``(n, r)`` where `n` is the number of samples and `p`, `q`, and `r`
        are the number of dimensions. Alternatively, ``x``, ``y`` and ``z``
        can be distance matrices where the shapes must be ``(n, n)``. ``z``
        is an optional ndarray.
    metric : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``metric`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``x``, ``y``, and ``z`` are already
        distance matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguements to send to your custom
        function.
    workers : int, default: 1
        The number of cores to parallelize the p-value computation over.
        Supply ``-1`` to use all cores available to the Process.
    **kwargs
        Arbitrary keyword arguments provided to
        :func:`sklearn.metrics.pairwise_distances` or a
        custom distance function.

    Returns
    -------
    distx, disty, distz : ndarray of float
        Distance matrices based on the metric provided by the user. ``distz`` is
        only returned if ``z`` is provided.
    """
    if not metric:
        metric = "precomputed"
    if callable(metric):
        distx = metric(x, **kwargs)
        disty = metric(y, **kwargs)
        distz = None if z is None else metric(z, **kwargs)
        _check_distmat(
            distx,
            disty,
            distz,
        )  # verify whether matrix is correct, built into sklearn func
    else:
        distx = pairwise_distances(x, metric=metric, n_jobs=workers, **kwargs)
        disty = pairwise_distances(y, metric=metric, n_jobs=workers, **kwargs)
        distz = (
            None
            if z is None
            else pairwise_distances(z, metric=metric, n_jobs=workers, **kwargs)
        )

    if z is None:
        return distx, disty
    else:
        return distx, disty, distz


def check_perm_blocks(perm_blocks):
    # Checks generic properties of perm_blocks
    if perm_blocks is None:
        return None
    elif isinstance(perm_blocks, list):
        perm_blocks = np.asarray(perm_blocks)
    elif not isinstance(perm_blocks, np.ndarray):
        raise TypeError("perm_blocks must be an ndarray or list")
    if perm_blocks.ndim == 1:
        perm_blocks = perm_blocks[:, np.newaxis]
    elif perm_blocks.ndim > 2:
        raise ValueError("perm_blocks must be of at most dimension 2")

    return perm_blocks


def check_perm_blocks_dim(perm_blocks, y):
    if not perm_blocks.shape[0] == y.shape[0]:
        raise ValueError("perm_bocks first dimension must be same length as y")


def check_perm_block(perm_block):
    # checks a hierarchy level of perm_blocks for proper exchangeability
    if not isinstance(perm_block[0], int):
        unique, perm_blocks, counts = np.unique(
            perm_block, return_counts=True, return_inverse=True
        )
        pos_counts = counts
    else:
        unique, counts = np.unique(perm_block, return_counts=True)
        pos_counts = [c for c, u in zip(counts, unique) if u >= 0]
    if len(set(pos_counts)) > 1:
        raise ValueError(
            f"Exchangeable hiearchy has groups with {min(pos_counts)} to \
                {max(pos_counts)} elements"
        )

    return perm_block


class _PermNode(object):
    """Helper class for nodes in _PermTree."""

    def __init__(self, parent, label=None, index=None):
        self.children = []
        self.parent = parent
        self.label = label
        self.index = index

    def get_leaf_indices(self):
        if len(self.children) == 0:
            return [self.index]
        else:
            indices = []
            for child in self.children:
                indices += child.get_leaf_indices()
            return indices

    def add_child(self, child):
        self.children.append(child)

    def get_children(self):
        return self.children


class _PermTree(object):
    """Tree representation of dependencies for restricted permutations"""

    def __init__(self, perm_blocks):
        perm_blocks = check_perm_blocks(perm_blocks)
        self.root = _PermNode(None)
        self._add_levels(self.root, perm_blocks, np.arange(perm_blocks.shape[0]))
        indices = self.root.get_leaf_indices()
        self._index_order = np.argsort(indices)

    def _add_levels(self, root: _PermNode, perm_blocks, indices):
        # Add new child node for each unique label, then recurse or end
        if perm_blocks.shape[1] == 0:
            for idx in indices:
                child_node = _PermNode(parent=root, label=1, index=idx)
                root.add_child(child_node)
        else:
            perm_block = check_perm_block(perm_blocks[:, 0])
            for label in np.unique(perm_block):
                idxs = np.where(perm_block == label)[0]
                child_node = _PermNode(parent=root, label=label)
                root.add_child(child_node)
                self._add_levels(child_node, perm_blocks[idxs, 1:], indices[idxs])

    def _permute_level(self, node, rng=None):
        if rng is None:
            rng = np.random
        if len(node.get_children()) == 0:
            return [node.index]
        else:
            indices, labels = zip(
                *[
                    (self._permute_level(child), child.label)
                    for child in node.get_children()
                ]
            )
            shuffle_children = [i for i, label in enumerate(labels) if label >= 0]
            indices = np.asarray(indices)
            if len(shuffle_children) > 1:
                indices[shuffle_children] = indices[rng.permutation(shuffle_children)]
            return np.concatenate(indices)

    def permute_indices(self, rng=None):
        return self._permute_level(self.root, rng)[self._index_order]

    def original_indices(self):
        return np.arange(len(self._index_order))


# permutation group shuffling class
class _PermGroups(object):
    """Helper function to calculate parallel p-value."""

    def __init__(self, y, perm_blocks=None):
        self.n = y.shape[0]
        if perm_blocks is None:
            self.perm_tree = None
        else:
            self.perm_tree = _PermTree(perm_blocks)

    def __call__(self, rng=None):
        rng = check_random_state(rng)
        if self.perm_tree is None:
            order = rng.permutation(self.n)
        else:
            order = self.perm_tree.permute_indices(rng)

        return order


# p-value computation
def _perm_stat(
    calc_stat, x, y, z=None, is_distsim=True, permuter=None, random_state=None
):
    """Permute the test statistic"""
    rng = check_random_state(random_state)
    if permuter is None:
        order = rng.permutation(y.shape[0])
    else:
        order = permuter(rng=rng)

    if is_distsim:
        permy = y[order][:, order]
    else:
        permy = y[order]

    if z is not None:
        perm_stat = calc_stat(x, permy, z)
    else:
        perm_stat = calc_stat(x, permy)

    return perm_stat


def perm_test(
    calc_stat,
    x,
    y,
    z=None,
    reps=1000,
    workers=1,
    is_distsim=True,
    perm_blocks=None,
    random_state=None,
    permuter=None,
):
    """
    Permutation test for the p-value of a nonparametric test.

    This process is completed by first randomly permuting :math:`y` to estimate the null
    distribution and then calculating the probability of observing a test
    statistic, under the null, at least as extreme as the observed test
    statistic.

    Parameters
    ----------
    calc_stat : callable
        The method used to calculate the test statistic (must use hyppo API).
    x,y,z : ndarray of float
        Input data matrices. ``x``, ``y`` and ``z`` must have the same number of
        samples. That is, the shapes must be ``(n, p)``, ``(n, q)``, ``(n, r)``where
        `n` is the number of samples and `p`, `q` , and `r` are the number of
        dimensions. Alternatively, ``x`` and ``y`` can be distance or similarity
        matrices, and ``z`` must be a similarity matrix where the shapes
        must be ``(n, n)``. ``z`` is an optional matrix only used for conditional
        independence testing.
    reps : int, default: 1000
        The number of replications used to estimate the null distribution
        when using the permutation test used to calculate the p-value.
    workers : int, default: 1
        The number of cores to parallelize the p-value computation over.
        Supply ``-1`` to use all cores available to the Process.
    is_distsim : bool, default: True
        Whether or not ``x`` and ``y`` are distance or similarity matrices.
    perm_blocks : ndarray, default: None
        Defines blocks of exchangeable samples during the permutation test.
        If ``None``, all samples can be permuted with one another. Requires `n`
        rows. Constructs a tree graph with all samples initially at
        the root node. Each column partitions samples from the same leaf with
        shared column label into a child of that leaf. During the permutation
        test, samples within the same final leaf node are exchangeable
        and blocks of samples with a common parent node are exchangeable. If a
        column value is negative, the resulting block is unexchangeable.
    permuter : callable, default: None
        Defines a custom permutation function. If None, the default permutation
        test is used.

    Returns
    -------
    stat : float
        The computed test statistic.
    pvalue : float
        The computed p-value.
    null_dist : list of float
        The approximated null distribution of shape ``(reps,)``.
    """

    # calculate observed test statistic
    if z is None:
        stat = calc_stat(x, y)
    else:
        stat = calc_stat(x, y, z)

    # make RandomState seeded array
    if random_state is not None:
        rng = check_random_state(random_state)
        random_state = rng.randint(np.iinfo(np.int32).max, size=reps)

    # make random array
    else:
        random_state = np.random.randint(np.iinfo(np.int32).max, size=reps)

    # calculate null distribution
    if not callable(permuter):
        permuter = _PermGroups(y, perm_blocks)

    null_dist = np.array(
        Parallel(n_jobs=workers)(
            [
                delayed(_perm_stat)(calc_stat, x, y, z, is_distsim, permuter, rng)
                for rng in random_state
            ]
        )
    )
    pvalue = (1 + (null_dist >= stat).sum()) / (1 + reps)

    return stat, pvalue, null_dist


def _multi_perm_stat(calc_stat, *args):
    """Permute every entry and calculate test statistic"""
    # permute each row
    comb_matrix = np.concatenate(args, axis=1)
    perm_matrix = np.zeros(np.shape(comb_matrix))
    for j in range(comb_matrix.shape[1]):
        order = np.random.permutation(comb_matrix.shape[0])
        perm_matrix[:, j] = comb_matrix[order, j]
    perm_args = tuple(np.split(perm_matrix, len(args), axis=1))

    # calculate test statistic using permuted matrices
    perm_stat = calc_stat(*perm_args)

    return perm_stat


def multi_perm_test(calc_stat, *args, reps=1000, workers=1):
    """
    Permutation test for the p-value of a nonparametric test with multiple variables.

    Parameters
    ----------
    calc_stat : callable
        The method used to calculate the test statistic (must use hyppo API).
    *args : ndarray
        Variable length input data matrices. All inputs must have the same
        number of samples. That is, the shapes must be ``(n, p)``, ``(n, q)``,
        etc., where `n` is the number of samples and `p` and `q` are the
        number of dimensions.
    reps : int, default: 1000
        The number of replications used to estimate the null distribution
        when using the permutation test used to calculate the p-value.
    workers : int, default: 1
        The number of cores to parallelize the p-value computation over.
        Supply ``-1`` to use all cores available to the Process.

    Returns
    -------
    stat : float
        The computed test statistic.
    pvalue : float
        The computed p-value.
    null_dist : list of float
        The approximated null distribution of shape ``(reps,)``.
    """
    # calculate observed test statistic
    stat = calc_stat(*args)

    # calculate null distribution
    null_dist = np.array(
        Parallel(n_jobs=workers)(
            [delayed(_multi_perm_stat)(calc_stat, *args) for _ in range(reps)]
        )
    )
    pvalue = (1 + (null_dist >= stat).sum()) / (1 + reps)

    return stat, pvalue, null_dist


def chi2_approx(calc_stat, x, y):
    """
    Fast chi-squared approximation for the p-value.

    In the case of distance and kernel methods, Dcorr (and by extension Hsic
    :footcite:p:`shenExactEquivalenceDistance2020`)
    can be approximated via a chi-squared distribution
    :footcite:p:`shenChiSquareTestDistance2021`.
    This approximation is also applicable for the nonparametric MANOVA via
    independence testing method in our package
    :footcite:p:`pandaNonparMANOVAIndependence2021`.

    Parameters
    ----------
    calc_stat : callable
        The method used to calculate the test statistic (must use hyppo API).
    x,y : ndarray of float
        Input data matrices. ``x`` and ``y`` must have the same number of
        samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
        `n` is the number of samples and `p` and `q` are the number of
        dimensions. Alternatively, ``x`` and ``y`` can be distance or similarity
        matrices,
        where the shapes must both be ``(n, n)``.

    Returns
    -------
    stat : float
        The computed test statistic.
    pvalue : float
        The computed p-value.

    References
    ----------
    .. footbibliography::
    """
    n = x.shape[0]
    stat = calc_stat(x, y)
    pvalue = chi2.sf(stat * n + 1, 1)

    return stat, pvalue



import autograd.numpy as np
from numpy import testing
from ..kernel import KGauss
from ...tools import compute_kern, multi_compute_kern

import pytest
from numpy.random import default_rng


class TestKGauss:
    @pytest.mark.parametrize("n", [10])
    @pytest.mark.parametrize("d", [3])
    def test_basic(self, n, d):
        # sample
        rng = default_rng(29)
        X = rng.standard_normal(size=(n, d)) * 3
        k = KGauss(sigma2=1)
        K = k.eval(X, X)  # tests usage of compute_kern

        testing.assert_almost_equal(K.shape, (n, n))

    @pytest.mark.parametrize("n", [10])
    @pytest.mark.parametrize("d", [1, 3])
    def test_gradX_y(self, n, d):
        rng = default_rng(10)
        y = rng.standard_normal(size=d) * 2
        X = rng.random(size=(n, d)) * 3
        Y = rng.standard_normal(size=(n, d)) * 2

        sigma2 = 1.3
        k = KGauss(sigma2=sigma2)
        # n x d
        G = k.gradX_y(X, y)
        k.gradX_Y(X, Y, dim=d - 1)
        k.gradXY_sum(X, Y)
        k.pair_gradX_Y(X, Y)
        k.pair_gradXY_sum(X, Y)
        k.pair_eval(X, Y)
        # check correctness
        K = k.eval(X, y[np.newaxis, :])
        modG = -K / sigma2 * (X - y)

        testing.assert_equal(G.shape, modG.shape)
        testing.assert_almost_equal(G, modG)

    @pytest.mark.parametrize("n", [11])
    @pytest.mark.parametrize("d", [3, 1])
    def test_gradXY_sum(self, n, d):
        rng = default_rng(12)
        X = rng.standard_normal(size=(n, d))
        sigma2 = 1.4
        k = KGauss(sigma2=sigma2)

        # n x n
        modG = np.zeros((n, n))
        K = k.eval(X, X)
        for i in range(n):
            for j in range(n):
                diffi2 = np.sum((X[i, :] - X[j, :]) ** 2)
                modG[i, j] = K[i, j] / sigma2 * (d - diffi2 / sigma2)

        # check correctness of function that calculates sum of the gradient across x_i and y_i in Y
        G = k.gradXY_sum(X, X)

        testing.assert_equal(G.shape, modG.shape)
        testing.assert_almost_equal(G, modG)



import numpy as np
from sklearn.utils import check_random_state

from .indep_sim import _CheckInputs


def indep_normal(n, p=1, random_state=None):
    r"""
    Independent standard normal distributions.

    :math:`(X, Y, Z) \in \mathbb{R} \times \mathbb{R} \times \mathbb{R}`:

    .. math::
        X, Y, Z &\sim N(0, 1)

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        Ignored.

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    rng = check_random_state(random_state)

    mean = np.array([0, 0, 0])
    cov = np.eye(3)

    x, y, z = rng.multivariate_normal(mean, cov, size=n).T

    return x, y, z


def indep_lognormal(n, p=1, random_state=None):
    r"""
    Independent lognormal and normal distributions.

    :math:`(X, Y, Z) \in \mathbb{R} \times \mathbb{R} \times \mathbb{R}`:
    .. math::
        X &\sim \text{log} N(0, 1)\\
        Y, Z &\sim N(0, 1)

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    x, y, z = indep_normal(n=n, p=p, random_state=random_state)
    x = np.exp(x)

    return x, y, z


def indep_binomial(n, p=1, random_state=None):
    r"""
    Independent binomial distributions.

    :math:`(X, Y, Z) \in \mathbb{R} \times \mathbb{R} \times \mathbb{R}^p`:
    .. math::
        X, Y, Z_i &\sim \text{Binom}(10, 0.5) \\
        Z &= (Z_1, Z_2, \ldots, Z_p)

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions for conditioning variable ``Z``.

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x`` and ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    rng = check_random_state(random_state)

    size = (n, 1)

    z = rng.binomial(10, 0.5, size=(n, p))
    x = rng.binomial(10, 0.5, size=size)
    y = rng.binomial(10, 0.5, size=size)

    return x, y, z


def cond_indep_normal(n, p=1, random_state=None):
    r"""
    Conditionally independent normal distributions.

    :math:`(X, Y, Z) \in \mathbb{R} \times \mathbb{R} \times \mathbb{R}`:
    .. math::
        \mu &= (0, 0, 0)\\
        \Sigma &= \begin{bmatrix}
        1 & 0.36 & 0.6 \\
        0.36 & 1 & 0.6 \\
        0.6 & 0.6 & 1
        \end{bmatrix}\\
        (X, Y, Z) &\sim MVN(\mu, \Sigma)

    The conditional covariance matrix is given by:
    .. math::
        \Sigma(X, Y | Z) &= \begin{bmatrix}
        0.64 & 0\\
        0 & 0.64
        \end{bmatrix}

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        Ignored.

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    rng = check_random_state(random_state)

    mean = np.array([0, 0, 0])
    cov = np.array([[1, 0.36, 0.6], [0.36, 1, 0.6], [0.6, 0.6, 1]])

    x, y, z = rng.multivariate_normal(mean, cov, size=n).T

    return x, y, z


def cond_indep_lognormal(n, p=1, random_state=None):
    r"""
    Conditionally independent lognormal and normal distributions.

    :math:`(X, Y, Z) \in \mathbb{R} \times \mathbb{R} \times \mathbb{R}`:

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        Ignored.

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    rng = check_random_state(random_state)

    mean = np.array([0, 0, 0])
    cov = np.array([[1, 0.36, 0.6], [0.36, 1, 0.6], [0.6, 0.6, 1]])

    x, y, z = rng.multivariate_normal(mean, cov, size=n).T
    x = np.exp(x)

    return x, y, z


def cond_indep_normal_nonlinear(n, p=1, random_state=None):
    r"""
    Conditionally independent normal distributions. Example 3 from :footcite:p:`wang2015conditional`.

    :math:`(X, Y, Z) \in \mathbb{R} \times \mathbb{R} \times \mathbb{R}`:
    .. math::
        X_1, Y_1, Z &\sim N(0, 1) \\
        Z_1 &= 0.5 \left( \frac{Z^3}{7} + \frac{Z}{2} \right) \\
        Z_2 &= \frac{Z^3}{2} + \frac{Z}{3} \\
        X_2 &= Z_1 + \tanh(X_1) \\
        X &= X_2 + \frac{X_2^3}{3}\\
        Y_2 &= Z_2 + Y_1 \\
        Y &= Y_2 + \frac{Y_2^3}{3}

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    rng = check_random_state(random_state)

    size = (n, 1)
    z = rng.normal(size=size)
    z1 = 0.5 * (np.power(z, 3) / 7 + z / 2)
    z2 = (np.power(z, 3) / 2 + z) / 3

    x1 = rng.normal(size=size)
    x2 = z1 + np.tanh(x1)
    x = x2 + np.power(x2, 3) / 3

    y1 = rng.normal(size=size)
    y2 = z2 + y1
    y = y2 + np.tanh(y2 / 3)

    return x, y, z


def cond_indep_binomial(n, p=1, random_state=None):
    r"""
    Conditionally independent binomial distributions.

    :math:`(X, Y, Z) \in \mathbb{R} \times \mathbb{R} \times \mathbb{R}^p`:

    .. math::
        X_1, Y_1, Z_i &\sim \text{Binom}(10, 0.5) \\
        Z &= (Z_1, Z_2, \ldots, Z_p) \\
        X &= X_1 + Z_1 + Z_2 + \cdots + Z_p \\
        Y &= Y_1 + Z_1 + Z_2 + \cdots + Z_p

    Examples 2 and 4 from :footcite:p:`wang2015conditional`.

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions for conditioning variable ``Z``.

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    rng = check_random_state(random_state)

    size = (n, 1)

    z = rng.binomial(10, 0.5, size=(n, p))
    x = rng.binomial(10, 0.5, size=size) + z.sum(axis=1, keepdims=True)
    y = rng.binomial(10, 0.5, size=size) + z.sum(axis=1, keepdims=True)

    return x, y, z


def correlated_binomial(n, p=1, random_state=None):
    r"""
    Conditionally dependent binomial distributions. Examples 6 from:footcite:p:`wang2015conditional`.

    :math:`(X, Y, Z) \in \mathbb{R} \times \mathbb{R} \times \mathbb{R}`:

    .. math::
        X_1, Z &\sim \text{Binom}(10, 0.5) \\
        X &= X_1 + Z \\
        Y &= (X_1 - 5)^4 + Z


    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions for conditioning variable ``Z``.

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    rng = check_random_state(random_state)

    x1, z = rng.binomial(10, 0.5, size=(2, n))

    x = x1 + z
    y = (x1 - 5) ** 4 + z

    return x, y, z


def correlated_normal(n, p=1, random_state=None):
    r"""
    Conditionally dependent normal distributions. Example 5 from :footcite:p:`wang2015conditional`

    :math:`(X, Y, Z) \in \mathbb{R} \times \mathbb{R} \times \mathbb{R}`:
    .. math::
        \mu &= (0, 0, 0)\\
        \Sigma &= \begin{bmatrix}
        1 & 0.7 & 0.6\\
        0.7 & 1 & 0.6\\
        0.6 & 0.6 & 1
        \end{bmatrix}\\
        (X, Y, Z) &\sim MVN(\mu, \Sigma)

    The conditional covariance matrix is given by:
    .. math::
        \Sigma(X, Y | Z) &= \begin{bmatrix}
        0.64 & 0.34\\
        0.34 & 0.64
        \end{bmatrix}

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    rng = check_random_state(random_state)

    mean = np.array([0, 0, 0])
    cov = np.array([[1, 0.7, 0.6], [0.7, 1, 0.6], [0.6, 0.6, 1]])

    x, y, z = rng.multivariate_normal(mean, cov, size=n).T

    return x, y, z


def correlated_normal_nonliear(n, p=1, random_state=None):
    r"""
    Conditionally dependent normal distributions with nonlinear dependence.
    Example 7 from :footcite:p:`wang2015conditional`

    :math:`(X, Y, Z) \in \mathbb{R} \times \mathbb{R} \times \mathbb{R}`:
    .. math::
        X_1, Y_1, Z, \epsilon &\sim N(0, 1) \\
        Z_1 &= 0.5(Z^3/7 + Z/2) \\
        Z_2 &= (Z^3/2 + Z)/3 \\
        X_2 &= Z_1 + \tanh(X_1) \\
        X_3 &= X_2 + X_2^3 / 3 \\
        Y_2 &= Z_2 + Y_1\\
        Y_3 &= Y_2 + \tanh(Y_2 / 3) \\
        X &= X_3 + \cosh\epsilon \\
        Y &= Y_3 + \cosh\epsilon^2

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    rng = check_random_state(random_state)

    x1, y1, z, eps = rng.normal(size=(n, 4)).T

    z1 = 0.5 * (z**3 / 7 + z / 2)
    z2 = (z**3 / 2 + z) / 3

    x2 = z1 + np.tanh(x1)
    x3 = x2 + x2**3 / 3

    y2 = z2 + y1
    y3 = y2 + np.tanh(y2 / 3)

    x = x3 + np.cosh(eps)
    y = y3 + np.cosh(eps**2)

    return x, y, z


def correlated_lognormal(n, p=1, random_state=None):
    r"""
    Example 5 from :footcite:p:`szekelyPartialDistanceCorrelation2014a`
    :math:`(X, Y, Z) \in \mathbb{R} \times \mathbb{R} \times \mathbb{R}`:

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    x, y, z = correlated_normal(n, p, random_state)
    x = np.exp(x)

    return x, y, z


def correlated_t_linear(n, p=4, random_state=None):
    r"""
    Conditionally dependent t-distributed data with linear dependence.
    Example 9 from :footcite:p:`wang2015conditional`

    :math:`(X, Y, Z) \in \mathbb{R}^p \times \mathbb{R} \times \mathbb{R}`:
    .. math::
        Z_1, Z_2, \cdots, Z_p, Z_{p+1}, Z_{p+2}  &\sim t(1)\\
        X &= (Z_1, Z_2, Z_{p-1}, Z_p + Z_{p+1})\\
        Y &= Z_1 + Z_2 + \cdots + Z_p + Z_{p+1} + Z_{p+2}\\
        Z &= Z_{p+1}


    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimentions for variable ``x``.

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    rng = check_random_state(random_state)

    z_i = rng.standard_t(df=1, size=(n, p + 2))

    x = z_i[:, :p]
    x[:, -1] += z_i[:, p]
    y = z_i.sum(axis=1)
    z = z_i[:, p]

    return x, y, z


def correlated_t_quadratic(n, p=10, random_state=None):
    r"""
    Conditionally dependent t-distributed data with quadratic dependence.
    Example 10 from :footcite:p:`wang2015conditional`

    :math:`(X, Y, Z) \in \mathbb{R}^10 \times \mathbb{R}^2 \times \mathbb{R}`:
    .. math::
        Z_1, Z_2, \cdots, Z_{13}  &\sim t(1)\\
        X_i &= Z_i, i = 1, 2, \cdots, 9\\
        X_{10} &= Z_{10} + Z_{11}\\
        Y_1 &= Z_1 Z_2 + Z_3 Z_4 + Z_5 Z_{11} + Z_{12}\\
        Y_2 &= Z_6 Z_7 + Z_8 Z_9 + Z_{10} Z_{11} + Z_{13}\\
        Z &= Z_{11}

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    rng = check_random_state(random_state)

    z = rng.standard_t(df=1, size=(13, n))

    x = z[:11]
    x[9] += z[10]
    y1 = z[0] * z[1] + z[2] * z[3] + z[4] * z[10] + z[11]
    y2 = z[5] * z[6] + z[7] * z[8] + z[9] * z[10] + z[12]
    y = np.array([y1, y2])
    z = z[10]

    return x.T, y.T, z


def correlated_t_nonlinear(n, p=4, random_state=None):
    r"""
    Conditionally dependent t-distributed data with nonlinear dependence.
    Example 11 from :footcite:p:`wang2015conditional`

    :math:`(X, Y, Z) \in \mathbb{R}^4 \times \mathbb{R}^2 \times \mathbb{R}^2`:
    .. math::
        Z_1, \ldots, Z_4 &\sim t(2)\\
        Y_1 &= \sin(Z_1) + \cos(Z_2) + Z_3^2 + Z_4^2\\
        Y_2 &= Z_1^2 + Z_2^2 + Z_3 + Z_4\\
        X &= (Z_1, Z_2, Z_3, Z_4)\\
        Y &= (Y_1, Y_2)\\
        Z &= (Z_1, Z_2)

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).

    Returns
    -------
    x,y,z : ndarray of float
        Simulated data matrices. ``x``, ``y``, and ``z``.

    References
    ----------
    .. footbibliography::
    """
    check_in = _CheckInputs(n, p=p)
    check_in()

    rng = check_random_state(random_state)

    z = rng.standard_t(df=2, size=(4, n))
    y1 = np.sin(z[0]) + np.cos(z[1]) + z[2] ** 2 + z[3] ** 2
    y2 = z[0] ** 2 + z[1] ** 2 + z[2] + z[3]
    y = np.array([y1, y2])

    return z.T, y.T, z[:2].T


COND_SIMULATIONS = {
    "independent_normal": indep_normal,
    "independent_lognormal": indep_lognormal,
    "independent_binomial": indep_binomial,
    "cond_independent_normal": cond_indep_normal,
    "cond_independent_lognormal": cond_indep_lognormal,
    "cond_independent_binomial": cond_indep_binomial,
    "cond_independent_normal_nonlinear": cond_indep_normal_nonlinear,
    "correlated_normal": correlated_normal,
    "correlated_lognormal": correlated_lognormal,
    "correlated_binomial": correlated_binomial,
    "correlated_normal_nonliear": correlated_normal_nonliear,
    "correlated_t_linear": correlated_t_linear,
    "correlated_t_nonlinear": correlated_t_nonlinear,
    "correlated_t_quadratic": correlated_t_quadratic,
}


def condi_indep_sim(sim, n, p, random_state=None, **kwargs):
    r"""
    Conditional independence simulation generator.

    Takes a simulation and the required parameters, and outputs the simulated
    data matrices.

    Parameters
    ----------
    sim : str
        The name of the simulation (from the :mod:`hyppo.tools` module) that is to be
        rotated.
    n : int
        The number of samples desired by the simulation (>= 5).
    p : int
        The number of dimensions desired by the simulation (>= 1).
    **kwargs
        Additional keyword arguements for the desired simulation.

    Returns
    -------
    x,y, z : ndarray of float
        Simulated data matrices.
    """
    if sim not in COND_SIMULATIONS.keys():
        raise ValueError(
            "sim_name must be one of the following: {}".format(
                list(COND_SIMULATIONS.keys())
            )
        )
    else:
        sim = COND_SIMULATIONS[sim]

    x, y, z = sim(n, p, random_state, **kwargs)

    if x.ndim == 1:
        x = x[:, np.newaxis]
    if y.ndim == 1:
        y = y[:, np.newaxis]
    if z.ndim == 1:
        z = z[:, np.newaxis]

    return x, y, z



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from ...tools import linear, power
from .. import dHsic  # type: ignore


class TestdHsicStat:
    @pytest.mark.parametrize("n, obs_stat", [(100, 0.04561), (200, 0.03911)])
    @pytest.mark.parametrize("obs_pvalue", [1 / 1000])
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = linear(n, 1)
        stat, pvalue = dHsic(gamma=0.5).test(x, y)

        assert_almost_equal(stat, obs_stat, decimal=2)
        assert_almost_equal(pvalue, obs_pvalue, decimal=2)


class TestdHsicTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            "dhsic",
            sim_type="multi",
            sim="multimodal_independence",
            n=100,
            p=1,
            alpha=0.05,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)



import numpy as np
from scipy.stats import chi2
from ._utils import _CheckInputs
from hyppo.ksample.base import KSampleTest, KSampleTestOutput


class SmoothCFTest(KSampleTest):
    r"""
    Smooth Characteristic Function test statistic and p-value

    The Smooth Characteristic Function test is a two-sample test that uses
    differences in the smoothed (analytic) characteristic function of
    two data distributions in order to determine how different
    the two data are :footcite:p:`chwialkowski2015fast`.

    Parameters
    ----------
    num_randfreq: int
        Used to construct random array with size ``(p, q)`` where `p` is the number of
        dimensions of the data and `q` is the random frequency at which the
        test is performed. These are the random test points at which test occurs (see notes).

    Notes
    -----
    The test statistic takes on the following form:

    .. math::

        nW_n\Sigma_n^{-1}W_n

    As seen in the above formulation, this test-statistic takes the same form as
    the Hotelling :math:`T^2` statistic. However, the components are
    defined differently in this case. Given data sets
    X and Y, define the following as :math:`Z_i`, the vector of differences:

    .. math::

        Z_i = (k(X_i, T_1) - k(Y_i, T_1), \ldots,
        k(X_i, T_J) - k(Y_i, T_J)) \in \mathbb{R}^J

    The above is the vector of differences between kernels at test points, :math:`T_j`.
    This same formulation is used in the Mean Embedding Test.
    Moving forward, :math:`W_n` can be defined:

    .. math::

        W_n = \frac{1}{n} \sum_{i = 1}^n Z_i

    This leaves :math:`\Sigma_n`, the covariance matrix as:

    .. math::

        \Sigma_n = \frac{1}{n}ZZ^T

    In the specific case of the Smooth Characteristic function test,
    the vector of differences can be defined as follows:

    .. math::

        Z_i = (f(X_i)\sin(X_iT_1) - f(Y_i)\sin(Y_iT_1),
        f(X_i)\cos(X_iT_1) - f(Y_i)\cos(Y_iT_1),\cdots) \in \mathbb{R}^{2J}

    Once :math:`S_n` is calculated, a threshold :math:`r_{\alpha}` corresponding to the
    :math:`1 - \alpha` quantile of a Chi-squared distribution w/ J degrees of freedom
    is chosen. Null is rejected if :math:`S_n` is larger than this threshold.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, num_randfreq=5):

        self.num_randfreq = num_randfreq
        KSampleTest.__init__(self)

    def statistic(self, x, y, random_state):
        r"""
        Calculates the smooth CF test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            dimensions. That is, the shapes must be ``(n, p)`` and ``(m, p)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions.
        random_state: int
            Set random seed for generation of test points

        Returns
        -------
        stat : float
            The computed Smooth CF statistic.
        """
        _, p = np.shape(x)
        if random_state:
            np.random.seed(random_state)
        random_frequencies = np.random.randn(p, self.num_randfreq)

        x_smooth = np.exp(-np.linalg.norm(x, axis=1) ** 2 / 2).reshape(-1, 1)
        y_smooth = np.exp(-np.linalg.norm(y, axis=1) ** 2 / 2).reshape(-1, 1)

        x_mat = x.dot(random_frequencies)
        y_mat = y.dot(random_frequencies)

        x_smooth_cf = np.concatenate(
            (np.sin(x_mat) * x_smooth, np.cos(x_mat) * x_smooth), 1
        )
        y_smooth_cf = np.concatenate(
            (np.sin(y_mat) * y_smooth, np.cos(y_mat) * y_smooth), 1
        )

        difference = x_smooth_cf - y_smooth_cf
        return smooth_cf_distance(difference)

    def test(self, x, y, random_state=None):
        r"""
        Calculates the smooth CF test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            dimensions. That is, the shapes must be ``(n, p)`` and ``(m, p)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions.
        random_state: int
            Set random seed for generation of test points

        Returns
        -------
        stat : float
            The computed Smooth CF statistic.
        pvalue : float
            The computed smooth CF p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.ksample import SmoothCFTest
        >>> np.random.seed(1234)
        >>> x = np.random.randn(500, 10)
        >>> y = np.random.randn(500, 10)
        >>> stat, pvalue = SmoothCFTest().test(x, y, random_state=1234)
        >>> '%.2f, %.3f' % (stat, pvalue)
        '4.70, 0.910'
        """
        check_input = _CheckInputs(inputs=[x, y], indep_test=None)
        x, y = check_input()

        stat = self.statistic(x, y, random_state)
        pvalue = chi2.sf(stat, 2 * self.num_randfreq)
        self.stat = stat
        self.pvalue = pvalue

        return KSampleTestOutput(stat, pvalue)


def smooth_cf_distance(difference):
    r"""
    Calculates the Smooth CF test statistic using the vector of differences.

    Parameters
    ----------
    difference : ndarray of float
        The vector of differences which indicates distance between mean embeddings.
    num_randfeatures : int
        The number of test frequencies

    Returns
    -------
    stat : float
        The computed smooth CF statistic.
    """
    num_samples, _ = np.shape(difference)
    sigma = np.cov(np.transpose(difference))

    mu = np.mean(difference, 0)

    stat = num_samples * mu.dot(np.linalg.solve(sigma, np.transpose(mu)))

    return stat



import numpy as np
from sklearn.metrics import pairwise_distances

from ..tools import check_ndarray_xy, check_reps, contains_nan, convert_xy_float64


class _CheckInputs:
    """Checks inputs for discriminability tests"""

    def __init__(self, x, y, reps=None, is_dist=False, remove_isolates=True):
        self.x = x
        self.y = y
        self.reps = reps
        self.is_distance = is_dist
        self.remove_isolates = remove_isolates

    def __call__(self):
        if len(self.x) > 1:
            if self.x[0].shape[0] != self.x[1].shape[0]:
                msg = "The input matrices do not have the same number of rows."
                raise ValueError(msg)

        tmp_ = []
        for x1 in self.x:
            check_ndarray_xy(x1, self.y)
            contains_nan(x1)
            contains_nan(self.y)
            check_min_samples(x1)
            x1, self.y = convert_xy_float64(x1, self.y)
            tmp_.append(self._condition_input(x1))

        self.x = tmp_

        if self.reps:
            check_reps(self.reps)

        return self.x, self.y

    def _condition_input(self, x1):
        """Checks whether there is only one subject and removes
        isolates and calculate distance."""
        uniques, counts = np.unique(self.y, return_counts=True)

        if (counts != 1).sum() <= 1:
            msg = "You have passed a vector containing only a single unique sample id."
            raise ValueError(msg)

        if self.remove_isolates:
            idx = np.isin(self.y, uniques[counts != 1])
            self.y = self.y[idx]

            x1 = np.asarray(x1)
            if not self.is_distance:
                x1 = x1[idx]
            else:
                x1 = x1[np.ix_(idx, idx)]

        if not self.is_distance:
            x1 = pairwise_distances(x1, metric="euclidean")

        return x1


def check_min_samples(x1):
    """Check if the number of samples is at least 3"""
    nx = x1.shape[0]

    if nx <= 10:
        raise ValueError("Number of samples is too low")



import numpy as np
from scipy.stats import t

from ..tools import perm_test
from ._utils import _CheckInputs
from .base import ConditionalIndependenceTest, ConditionalIndependenceTestOutput


class PartialCorr(ConditionalIndependenceTest):
    r"""
    Conditional Pearson's correlation test.

    Partial correlation is a measure of the association between two univariate
    variables given a third univariate variable.

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already distance
        matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguements to send to your custom
        function.
    **kwargs
        Arbitrary keyword arguments for ``compute_distance``.

    Notes
    -----
    The statistic is computed as follows:

    .. math::
        r_{x, y ; z} = \frac{\rho_{xy} - \rho_{xz} \rho_{yz}}{\sqrt{(1 - \rho_{xz}^2)(1 - \rho_{yz}^2)}}

    where :math:`\rho_{xy}` is the Pearson correlation coefficient between :math:`x` and
    :math:`y`. The partial correlation test is implemented as a t-test footcite:p:`legendre2000`:.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, **kwargs):
        ConditionalIndependenceTest.__init__(self, **kwargs)

    def statistic(self, x, y, z):
        r"""
        Helper function that calculates the partial correlation test statistic.

        Parameters
        ----------
        x,y,z : ndarray of float
            Input data matrices. ``x``, ``y`` and ``z`` must have the same number
            of samples. That is, the shapes must be ``(n, p)``, ``(n, q)`` and
            ``(n, r)`` where `n` is the number of samples and `p`, `q`, and `r`
            are the number of dimensions. Alternatively, ``x`` and ``y`` can be
            distance matrices and ``z`` can be a similarity matrix where the
            shapes must be ``(n, n)``.

        Returns
        -------
        stat : float
            The computed partial correlation test statistic.
        """
        check_input = _CheckInputs(x, y, z, max_dims=1)
        x, y, z = check_input()

        corr = np.corrcoef(np.hstack([x, y, z]), rowvar=False)
        cov_xy, cov_xz, cov_yz = corr[np.triu_indices_from(corr, k=1)]

        if cov_xz <= 0 or cov_yz <= 0:
            stat = 0
        else:
            stat = (cov_xy - cov_xz * cov_yz) / np.sqrt(
                (1 - cov_xz**2) * (1 - cov_yz**2)
            )

        self.stat = stat
        return stat

    def test(
        self,
        x,
        y,
        z,
        reps=1000,
        workers=1,
        auto=True,
        perm_blocks=None,
        random_state=None,
    ):
        r"""
        Calculates the partial correlation test statistic and p-value.

        Parameters
        ----------
        x,y,z : ndarray of float
            Input data matrices. ``x``, ``y`` and ``z`` must have the same number
            of samples. That is, the shapes must be ``(n, 1)``, ``(n, 1)`` and
            ``(n, 1)`` where `n` is the number of samples and `p`, `q`, and `r`
            are the number of dimensions.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        auto : bool, default: True
            If True, the p-value is computed by t-distribution approximation.
            Parameters ``reps`` and ``workers`` are irrelevant in this case.
            Otherwise, :class:`hyppo.tools.perm_test` will be run.
        perm_blocks : None or ndarray, default: None
            Defines blocks of exchangeable samples during the permutation test.
            If None, all samples can be permuted with one another. Requires `n`
            rows. At each column, samples with matching column value are
            recursively partitioned into blocks of samples. Within each final
            block, samples are exchangeable. Blocks of samples from the same
            partition are also exchangeable between one another. If a column
            value is negative, that block is fixed and cannot be exchanged.
        random_state : int, default: None
            The random_state for permutation testing to be fixed for
            reproducibility.

        Returns
        -------
        stat : float
            The computed partial correlation test statistic.
        pvalue : float
            The computed partial correlation test p-value.
        """
        check_input = _CheckInputs(x, y, z, reps=reps, max_dims=1)
        x, y, z = check_input()

        if auto:  # run t-stat
            stat = self.statistic(x, y, z)

            n = x.shape[0]
            dof = n - 3
            tstat = stat * np.sqrt(dof) / np.sqrt(1 - stat**2)
            pvalue = 1 - t.cdf(np.abs(tstat), dof)

            self.stat = stat
            self.pvalue = pvalue
            self.null_dist = None
        else:  # run permutation
            stat, pvalue, null_dist = perm_test(
                self.statistic,
                x=x,
                y=y,
                z=z,
                reps=reps,
                workers=workers,
                is_distsim=False,
                perm_blocks=perm_blocks,
                random_state=random_state,
            )

            self.stat = stat
            self.pvalue = pvalue
            self.null_dist = null_dist

        return ConditionalIndependenceTestOutput(stat, pvalue)



from math import ceil

import numpy as np

from ..conditional import COND_INDEP_TESTS
from ..d_variate import MULTI_TESTS
from ..independence import INDEP_TESTS
from ..ksample import KSAMP_TESTS, KSample, k_sample_transform
from .indep_sim import indep_sim
from .ksample_sim import gaussian_3samp, rot_ksamp
from .conditional_indep_sim import condi_indep_sim

_ALL_SIMS = {
    "indep": indep_sim,
    "multi": indep_sim,
    "ksamp": rot_ksamp,
    "gauss": gaussian_3samp,
    "condi": condi_indep_sim,
}

_NONPERM_TESTS = {
    "dcorr": "fast",
    "hsic": "fast",
    "dhsic": "fast",
    "energy": "fast",
    "mmd": "fast",
    "disco": "fast",
    "manova": "analytical",
    "hotelling": "analytical",
    "kmerf": "fast",
}


def _sim_gen(sim_type, **kwargs):
    """
    Generate ``sims`` for the desired simulations.
    """
    if sim_type in ["indep", "ksamp"]:
        if (
            kwargs["sim"] in ["multiplicative_noise", "multimodal_independence"]
            and "noise" in kwargs.keys()
        ):
            kwargs.pop("noise")

    sims = _ALL_SIMS[sim_type](**kwargs)

    return sims


def _indep_perm_stat(test, sim_type, **kwargs):
    """
    Generates null and alternate distributions for the independence test.
    """
    x, y = _sim_gen(sim_type=sim_type, **kwargs)
    obs_stat = test.statistic(x, y)
    permy = np.random.permutation(y)
    perm_stat = test.statistic(x, permy)

    return obs_stat, perm_stat


def _ksamp_perm_stat(test, sim_type, **kwargs):
    """
    Generates null and alternate distributions for the k-sample test.
    """
    sims = _sim_gen(sim_type=sim_type, **kwargs)
    u, v = k_sample_transform(sims)
    obs_stat = test.statistic(u, v)
    permv = np.random.permutation(v)
    perm_stat = test.statistic(u, permv)

    return obs_stat, perm_stat


def _multi_perm_stat(test, sim_type, **kwargs):
    """
    Generates null and alternate distributions for the d_variate independence test.
    """
    x, y = _sim_gen(sim_type=sim_type, **kwargs)
    obs_stat = test.statistic(*(x, y))
    [permx, permy] = np.split(np.random.permutation(np.append(x, y)), 2)
    permx = permx.reshape(permx.shape[0], 1)
    permy = permy.reshape(permy.shape[0], 1)
    perm_stat = test.statistic(*(permx, permy))

    return obs_stat, perm_stat


def _conditional_perm_stat(test, sim_type, **kwargs):
    """
    Generates null and alternate distributions for the conditional independence test.
    """
    x, y, z = _sim_gen(sim_type=sim_type, **kwargs)
    obs_stat = test.statistic(x, y, z)

    # Separate if block for conditional distance correlation since it requires
    # special permutation
    if str(test).lower() == "cdcorr":
        probs = test._compute_kde(z)
        idx = test._permuter(probs)
        permy = y[idx]
        perm_stat = test.statistic(x, permy, z)
    else:
        permy = np.random.permutation(y)
        perm_stat = test.statistic(x, permy, z)

    return obs_stat, perm_stat


_PERM_STATS = {
    "indep": _indep_perm_stat,
    "ksamp": _ksamp_perm_stat,
    "multi": _multi_perm_stat,
    "condi": _conditional_perm_stat,
}


def _nonperm_pval(test, sim_type, **kwargs):
    """
    Generates fast  permutation pvalues
    """
    sims = _sim_gen(sim_type=sim_type, **kwargs)
    pvalue = test.test(*sims)[1]

    return pvalue


def power(test, sim_type, sim=None, n=100, alpha=0.05, reps=1000, auto=False, **kwargs):
    """
    Computes empircal power for hypothesis tests

    Parameters
    ----------
    test : str or list
        The name of the independence test (from the :mod:`hyppo.independence` module,
        the :mod:`hyppo.d_variate` module, or the :mod:`hyppo.conditional` module)
        that is to be tested. If MaxMargin, accepts list with first entry "MaxMargin"
        and second entry the name of another independence test.
        For :class:`hyppo.ksample.KSample` put the name of the independence test.
        For other tests in :mod:`hyppo.ksample` just use the name of the class.
    sim_type : "indep", "ksamp", "gauss", "multi", "condi"
        Type of power method to calculate. Depends on the type of ``sim``.
    sim : str, default: None
        The name of the independence simulation (from the :mod:`hyppo.tools` module).
        that is to be used. Set to ``None`` if using Gaussian simulation curve.
    n : int, default: 100
        The number of samples desired by the simulation (>= 5).
    alpha : float, default: 0.05
        The alpha level of the test.
    reps : int, default: 1000
        The number of replications used to estimate the null distribution
        when using the permutation test used to calculate the p-value.
    auto : bool, default: False
        Automatically uses fast approximation when `n` and size of array
        is greater than 20 or test has a non-permutation based p-value.
        If ``True``, and sample size is greater than 20, then
        :class:`hyppo.tools.chi2_approx` will be run. ``reps`` is
        irrelevant in this case.
        See documentation for ``test`` if this parameter applies.
    **kwargs
        Additional keyword arguements for ``sim``.

    Returns
    -------
    empirical_power : ndarray of float
        Estimated empirical power for the test.
    """
    if sim_type not in _ALL_SIMS.keys():
        raise ValueError(
            "sim_type {}, must be in {}".format(sim_type, list(_ALL_SIMS.keys()))
        )

    if type(test) is list:
        test_name = [t.lower() for t in test]
        if test_name[0] != "maxmargin" or test_name[1] not in INDEP_TESTS.keys():
            raise ValueError(
                "Test 1 must be Maximal Margin, currently {}; Test 2 must be an "
                "independence test, currently {}".format(*test)
            )
        test = INDEP_TESTS[test_name[0]](indep_test=test_name[1])
        test_name = test_name[1]
    else:
        test_name = test.lower()
        if test_name in INDEP_TESTS.keys():
            test = INDEP_TESTS[test_name]()
        elif test_name in KSAMP_TESTS.keys():
            test = KSAMP_TESTS[test_name]()
        elif test_name in MULTI_TESTS.keys():
            test = MULTI_TESTS[test_name]()
        elif test_name in COND_INDEP_TESTS.keys():
            test = COND_INDEP_TESTS[test_name]()
        else:
            raise ValueError(
                "Test {} not in {}".format(
                    test_name,
                    list(INDEP_TESTS.keys())
                    + list(KSAMP_TESTS.keys())
                    + list(MULTI_TESTS.keys())
                    + list(COND_INDEP_TESTS.keys()),
                )
            )

    kwargs["n"] = n
    perm_type = "indep"
    if sim_type in ["ksamp", "gauss"]:
        perm_type = "ksamp"
    if sim_type == "multi":
        perm_type = "multi"
    if sim_type == "condi":
        perm_type = "condi"
    if sim_type != "gauss":
        kwargs["sim"] = sim

    if test_name in _NONPERM_TESTS.keys() and (
        auto or _NONPERM_TESTS[test_name] == "analytical"
    ):
        if test_name in INDEP_TESTS.keys() and perm_type == "ksamp":
            test = KSample(test_name)
        if n < 20 and _NONPERM_TESTS[test_name] == "fast":
            empirical_power = np.nan
        else:
            pvals = np.array(
                [
                    _nonperm_pval(test=test, sim_type=sim_type, **kwargs)
                    for _ in range(reps)
                ]
            )
            empirical_power = (1 + (pvals <= alpha).sum()) / (1 + reps)
    else:
        alt_dist, null_dist = map(
            np.float64,
            zip(
                *[
                    _PERM_STATS[perm_type](test=test, sim_type=sim_type, **kwargs)
                    for _ in range(reps)
                ]
            ),
        )
        cutoff = np.sort(null_dist)[ceil(reps * (1 - alpha))]
        empirical_power = (1 + (alt_dist >= cutoff).sum()) / (1 + reps)

    return empirical_power



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from ...tools import cross_corr_ar, nonlinear_process
from .. import LjungBox


class TestLjungBox:
    def test_invalid_inputs(self):
        x = np.ones(5)
        y = np.ones(5)
        with pytest.raises(ValueError):
            LjungBox(max_lag=0).test(x, y)

    def test_statistic_test_functions(self):
        np.random.seed(1)

        n = 20
        x = np.arange(1, n + 1).reshape(n, 1)
        y = np.arange(1, n + 1).reshape(n, 1)

        stat = LjungBox(max_lag=1).statistic(x, y)[0]
        stat2 = LjungBox(max_lag=1).test(x, y)[0]

        assert stat == stat2

    def test_same_seq(self):
        np.random.seed(1)

        n = 20
        x = np.arange(1, n + 1).reshape(n, 1)
        y = np.arange(1, n + 1).reshape(n, 1)
        stat = LjungBox(max_lag=1).statistic(x, y)[0]
        assert_almost_equal(stat, 18.53, decimal=2)

        res = LjungBox(max_lag=1).test(x, y)
        assert_almost_equal(res[1], 0.0, decimal=2)

    def test_same_seq_permutations(self):
        np.random.seed(2)

        n = 20
        x = np.arange(1, n + 1).reshape(n, 1)
        y = np.arange(1, n + 1).reshape(n, 1)
        res = LjungBox(max_lag=1).test(x, y, auto=False)
        assert_almost_equal(res[1], 0.0, decimal=2)

    def test_linear(self):
        np.random.seed(3)

        x, y = cross_corr_ar(100, lag=1)
        _, pvalue, mgcx_dict = LjungBox(max_lag=5).test(x, y)
        opt_lag = mgcx_dict["opt_lag"]

        assert_almost_equal(pvalue, 1 / 1000, decimal=2)
        assert_almost_equal(opt_lag, 1, decimal=0)

    def test_nonlinear(self):
        np.random.seed(4)

        x, y = nonlinear_process(100, lag=1)
        _, pvalue, mgcx_dict = LjungBox(max_lag=1).test(x, y, auto=False)
        opt_lag = mgcx_dict["opt_lag"]

        assert_almost_equal(pvalue, 0.9, decimal=1)
        assert_almost_equal(opt_lag, 1, decimal=0)



from ..independence import INDEP_TESTS
from ._utils import _CheckInputs, k_sample_transform
from .base import KSampleTest


class KSample(KSampleTest):
    r"""
    Nonparametric `K`-Sample Testing test statistic and p-value.

    A *k*-sample test tests equality in distribution among groups. Groups
    can be of different sizes, but generally have the same dimensionality.
    There are not many non-parametric *k*-sample tests, but this version
    cleverly leverages the power of some of the implemented independence
    tests to test this equality of distribution.

    Parameters
    ----------
    indep_test : "CCA", "Dcorr", "HHG", "RV", "Hsic", "MGC", "KMERF", or list
        A string corresponding to the desired independence test from
        :mod:`hyppo.independence`. This is not case sensitive. If using ``"MaxMargin"``
        then this must be a list containing ``"MaxMargin"`` in the first index and
        another ``indep_test`` in the second index.
    compute_distkern : str, callable, or None, default: "euclidean" or "gaussian"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Alternatively, this function computes the kernel similarity among the
        samples within each data matrix.
        Valid strings for ``compute_kernel`` are, as defined in
        :func:`sklearn.metrics.pairwise.pairwise_kernels`,

            [``"additive_chi2"``, ``"chi2"``, ``"linear"``, ``"poly"``,
            ``"polynomial"``, ``"rbf"``,
            ``"laplacian"``, ``"sigmoid"``, ``"cosine"``]

        Note ``"rbf"`` and ``"gaussian"`` are the same metric.
    bias : bool, default: False
        Whether or not to use the biased or unbiased test statistics (for
        ``indep_test="Dcorr"`` and ``indep_test="Hsic"``).
    **kwargs
        Arbitrary keyword arguments for ``compute_distkern``.

    Notes
    -----
    .. note::
        This algorithm is currently under review at a peer-reviewed journal.

    The formulation for this implementation is as follows
    :footcite:p:`pandaNonparMANOVAIndependence2021`:

    The *k*-sample testing problem can be thought of as a generalization of
    the two sample testing problem. Define
    :math:`\{ u_i \stackrel{iid}{\sim} F_U,\ i = 1, ..., n \}` and
    :math:`\{ v_j \stackrel{iid}{\sim} F_V,\ j = 1, ..., m \}` as two groups
    of samples deriving from different distributions with the same
    dimensionality. Then, problem that we are testing is thus,

    .. math::

        H_0: F_U &= F_V \\
        H_A: F_U &\neq F_V

    The closely related independence testing problem can be generalized
    similarly: Given a set of paired data
    :math:`\{\left(x_i, y_i \right) \stackrel{iid}{\sim} F_{XY}, \ i = 1, ..., N\}`,
    the problem that we are testing is,

    .. math::

        H_0: F_{XY} &= F_X F_Y \\
        H_A: F_{XY} &\neq F_X F_Y

    By manipulating the inputs of the *k*-sample test, we can create
    concatenated versions of the inputs and another label matrix which are
    necessarily paired. Then, any nonparametric test can be performed on
    this data.

    Letting :math:`n = \sum_{i=1}^k n_i`, define new data matrices
    :math:`\mathbf{x}` and :math:`\mathbf{y}` such that,

    .. math::

        \begin{align*}
        \mathbf{x} &=
        \begin{bmatrix}
            \mathbf{u}_1 \\
            \vdots \\
            \mathbf{u}_k
        \end{bmatrix} \in \mathbb{R}^{n \times p} \\
        \mathbf{y} &=
        \begin{bmatrix}
            \mathbf{1}_{n_1 \times 1} & \mathbf{0}_{n_1 \times 1}
            & \ldots & \mathbf{0}_{n_1 \times 1} \\
            \mathbf{0}_{n_2 \times 1} & \mathbf{1}_{n_2 \times 1}
            & \ldots & \mathbf{0}_{n_2 \times 1} \\
            \vdots & \vdots & \ddots & \vdots \\
            \mathbf{0}_{n_k \times 1} & \mathbf{0}_{n_k \times 1}
            & \ldots & \mathbf{1}_{n_k \times 1} \\
        \end{bmatrix} \in \mathbb{R}^{n \times k}
        \end{align*}

    Additionally, in the two-sample case,

    .. math::

        \begin{align*}
        \mathbf{x} &=
        \begin{bmatrix}
            \mathbf{u}_1 \\
            \mathbf{u}_2
        \end{bmatrix} \in \mathbb{R}^{n \times p} \\
        \mathbf{y} &=
        \begin{bmatrix}
            \mathbf{0}_{n_1 \times 1} \\
            \mathbf{1}_{n_2 \times 1}
        \end{bmatrix} \in \mathbb{R}^n
        \end{align*}

    Given :math:`\mathbf{u}` and :math:`\mathbf{v}`$` as defined above,
    to perform a :math:`w`-way test where :math:`w < k`,

    .. math::

        \mathbf{y} =
        \begin{bmatrix}
            \mathbf{1}_{n_1 \times 1} & \mathbf{0}_{n_1 \times 1}
            & \ldots & \mathbf{1}_{n_1 \times 1} \\
            \mathbf{1}_{n_2 \times 1} & \mathbf{1}_{n_2 \times 1}
            & \ldots & \mathbf{0}_{n_2 \times 1} \\
            \vdots & \vdots & \ddots & \vdots \\
            \mathbf{0}_{n_k \times 1} & \mathbf{1}_{n_k \times 1}
            & \ldots & \mathbf{1}_{n_k \times 1} \\
        \end{bmatrix} \in \mathbb{R}^{n \times k}.

    where each row of :math:`\mathbf{y}` contains :math:`w`
    :math:`\mathbf{1}_{n_i}` elements. This leads to label matrix distances
    proportional to how many labels (ways) samples differ by, a hierarchy of distances
    between samples thought to be true if the null hypothesis is rejected.

    Performing a multilevel test involves constructing :math:`x` and :math:`y` using
    either of the methods above and then performing a block permutation
    :footcite:p:`winklerMultilevelBlockPermutation2015`.
    Essentially, the permutation is striated, where permutation is limited to be within
    a block of samples or between blocks of samples, but not both. This is done because
    the data is not freely exchangeable, so it is necessary to block the permutation to
    preserve the joint distribution :footcite:p:`winklerMultilevelBlockPermutation2015`.

    The p-value returned is calculated using a permutation test uses
    :meth:`hyppo.tools.perm_test`.
    The fast version of the test uses :meth:`hyppo.tools.chi2_approx`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, indep_test, compute_distkern="euclidean", bias=False, **kwargs):
        if type(indep_test) is list:
            indep_test = [test.lower() for test in indep_test]
            self.indep_test_name = indep_test[1]
        else:
            indep_test = indep_test.lower()
            if indep_test not in INDEP_TESTS.keys():
                raise ValueError(
                    "Test {} is not in {}".format(indep_test, INDEP_TESTS.keys())
                )
            if indep_test == "hsic" and compute_distkern == "euclidean":
                compute_distkern = "gaussian"
            self.indep_test_name = indep_test

        indep_kwargs = {
            "dcorr": {"bias": bias, "compute_distance": compute_distkern},
            "hsic": {"bias": bias, "compute_kernel": compute_distkern},
            "hhg": {"compute_distance": compute_distkern},
            "mgc": {"compute_distance": compute_distkern},
            "kmerf": {"forest": "classifier", "is_ksamp": k_sample_transform},
            "rv": {},
            "cca": {},
        }

        if type(indep_test) is list:
            if indep_test[0] == "maxmargin" and indep_test[1] in INDEP_TESTS.keys():
                if indep_test[1] == "hsic" and compute_distkern == "euclidean":
                    compute_distkern = "gaussian"
                indep_kwargs["maxmargin"] = {
                    "indep_test": indep_test[1],
                    "compute_distkern": compute_distkern,
                    "bias": bias,
                }
                indep_test = "maxmargin"
            else:
                raise ValueError(
                    "Test 1 must be Maximal Margin, currently {}; Test 2 must be an "
                    "independence test, currently {}".format(*indep_test)
                )

        self.indep_test = INDEP_TESTS[indep_test](**indep_kwargs[indep_test], **kwargs)

        KSampleTest.__init__(
            self, compute_distance=compute_distkern, bias=bias, **kwargs
        )

    def statistic(self, *args):
        r"""
        Calulates the *k*-sample test statistic.

        Parameters
        ----------
        *args : ndarray of float
            Variable length input data matrices. All inputs must have the same
            number of dimensions. That is, the shapes must be `(n, p)` and
            `(m, p)`, ... where `n`, `m`, ... are the number of samples and `p` is
            the number of dimensions.

        Returns
        -------
        stat : float
            The computed *k*-sample statistic.
        """
        inputs = list(args)
        if self.indep_test_name == "kmerf":
            u, v = k_sample_transform(inputs, test_type="rf")
        else:
            u, v = k_sample_transform(inputs)

        return self.indep_test.statistic(u, v)

    def test(self, *args, reps=1000, workers=1, auto=True, random_state=None):
        r"""
        Calculates the *k*-sample test statistic and p-value.

        Parameters
        ----------
        *args : ndarray of float
            Variable length input data matrices. All inputs must have the same
            number of dimensions. That is, the shapes must be `(n, p)` and
            `(m, p)`, ... where `n`, `m`, ... are the number of samples and `p` is
            the number of dimensions.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        auto : bool, default: True
            Only applies to ``"Dcorr"`` and ``"Hsic"``.
            Automatically uses fast approximation when `n` and size of array
            is greater than 20. If ``True``, and sample size is greater than 20, then
            :class:`hyppo.tools.chi2_approx` will be run. Parameters ``reps`` and
            ``workers`` are
            irrelevant in this case. Otherwise, :class:`hyppo.tools.perm_test` will be
            run.

        Returns
        -------
        stat : float
            The computed *k*-sample statistic.
        pvalue : float
            The computed *k*-sample p-value.
        dict
            A dictionary containing optional parameters for tests that return them.
            See the relevant test in :mod:`hyppo.independence`.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.ksample import KSample
        >>> x = np.arange(7)
        >>> y = x
        >>> z = np.arange(10)
        >>> stat, pvalue = KSample("Dcorr").test(x, y)
        >>> '%.3f, %.1f' % (stat, pvalue)
        '0.000, 1.0'
        """
        inputs = list(args)
        check_input = _CheckInputs(
            inputs=inputs,
            indep_test=self.indep_test_name,
        )
        inputs = check_input()
        if self.indep_test_name == "kmerf":
            u, v = k_sample_transform(inputs, test_type="rf")
        else:
            u, v = k_sample_transform(inputs)
        v = v.astype(int)

        kwargs = {}
        if self.indep_test_name in ["dcorr", "hsic"]:
            kwargs = {"auto": auto}
        elif self.indep_test_name in ["kmerf"]:
            kwargs = {"auto": auto}

        return self.indep_test.test(
            u, v, reps, workers, **kwargs, random_state=random_state
        )



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from ...tools import indep_normal, power
from .. import PartialCorr


class TestPcorrStat:
    @pytest.mark.parametrize("n", [100, 200])
    @pytest.mark.parametrize("obs_stat", [0.0])
    def test_linear_oned(self, n, obs_stat):
        np.random.seed(3)
        x, y, z = indep_normal(n, 1)
        stat1, pvalue1 = PartialCorr().test(x, y, z)
        stat2 = PartialCorr().statistic(x, y, z)

        assert_almost_equal(stat1, obs_stat, decimal=1)
        assert_almost_equal(stat2, obs_stat, decimal=1)

    @pytest.mark.parametrize("n", [100, 200])
    def test_rep(self, n):
        x, y, z = indep_normal(n, 1)
        stat1, pvalue1 = PartialCorr().test(
            x, y, z, random_state=2, auto=False, reps=1000
        )
        stat2, pvalue2 = PartialCorr().test(
            x, y, z, random_state=2, auto=False, reps=1000
        )

        assert stat1 == stat2
        assert pvalue1 == pvalue2


class TestPcorrTypeIError:
    def test_oned(self):
        np.random.seed(1)
        est_power = power(
            "PartialCorr",
            sim_type="condi",
            sim="independent_normal",
            n=100,
            p=1,
            alpha=0.05,
        )
        assert_almost_equal(0.05, est_power, decimal=2)



import numpy as np
from numpy.testing import assert_almost_equal, assert_raises

from .. import power


class TestPower:
    def test_power_perm(self):
        np.random.seed(123456789)
        est_power = power("RV", "indep", sim="linear", p=1)
        assert_almost_equal(est_power, 1.0, decimal=1)

    def test_power_fast(self):
        np.random.seed(123456789)
        est_power = power("Dcorr", "indep", sim="linear", p=1, auto=True)
        assert_almost_equal(est_power, 1.0, decimal=1)

    def test_power_fast_nless20(self):
        np.random.seed(123456789)
        est_power = power("Dcorr", "indep", sim="linear", n=19, p=1, auto=True)
        assert_almost_equal(est_power, np.nan, decimal=1)

    def test_ksamp(self):
        np.random.seed(123456789)
        est_power = power("Dcorr", "ksamp", sim="linear", p=1, auto=True)
        assert_almost_equal(est_power, 1.0, decimal=1)

    def test_gaussian(self):
        np.random.seed(123456789)
        est_power = power("Dcorr", "gauss", auto=True, case=2)
        assert_almost_equal(est_power, 1.0, decimal=1)

    def test_multi(self):
        np.random.seed(123456789)
        est_power = power(
            "dHsic",
            sim_type="multi",
            sim="multimodal_independence",
            n=100,
            p=1,
            auto=False,
        )
        assert_almost_equal(est_power, 0.05, decimal=1)

    def test_condi_power_perm(self):
        np.random.seed(123456789)
        est_power = power(
            "conditionaldcorr",
            sim_type="condi",
            sim="independent_normal",
            n=50,
            p=1,
            auto=False,
        )
        assert_almost_equal(est_power, 0.05, decimal=2)


class TestPowerErrorWarn:
    def test_power_nosim(self):
        np.random.seed(123456789)
        assert_raises(ValueError, power, "CCA", "abcd")

    def test_power_noindep(self):
        np.random.seed(123456789)
        assert_raises(ValueError, power, "abcd", "indep")

    def test_power_nomaxmargin(self):
        np.random.seed(123456789)
        assert_raises(ValueError, power, ["MaxMargin", "abcd"], "indep")



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from ...tools import power, rot_ksamp
from .. import MMD


class TestMMD:
    @pytest.mark.parametrize(
        "n, obs_stat, obs_pvalue",
        [(200, 4.28e-7, 0.001), (100, 8.24e-5, 0.001)],
    )
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = rot_ksamp("linear", n, 1, k=2)
        stat, pvalue = MMD().test(x, y, auto=False)

        assert_almost_equal(stat, obs_stat, decimal=1)
        assert_almost_equal(pvalue, obs_pvalue, decimal=1)

    @pytest.mark.parametrize(
        "n, obs_stat, obs_pvalue",
        [(100, 8.24e-5, 0.001)],
    )
    def test_rep(self, n, obs_stat, obs_pvalue):
        x, y = rot_ksamp("linear", n, 1, k=2)
        stat, pvalue = MMD().test(x, y, auto=False, random_state=2)
        stat2, pvalue2 = MMD().test(x, y, auto=False, random_state=2)

        assert stat == stat2
        assert pvalue == pvalue2


class TestMMDTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            "MMD",
            sim_type="ksamp",
            sim="multimodal_independence",
            k=2,
            n=100,
            p=1,
            alpha=0.05,
            auto=True,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)



import numpy as np
import pytest
from numpy.testing import assert_raises
from sklearn.ensemble import RandomForestRegressor

from ...tools.common import _check_kernmat
from .._utils import _CheckInputs, sim_matrix


class TestErrorWarn:
    """Tests errors and warnings."""

    def test_error_notndarray(self):
        # raises error if x or y is not a ndarray
        x = np.arange(20)
        y = [5] * 20
        assert_raises(TypeError, _CheckInputs(x, y))
        assert_raises(TypeError, _CheckInputs(y, x))

    def test_error_shape(self):
        # raises error if number of samples different (n)
        x = np.arange(100).reshape(25, 4)
        y = x.reshape(10, 10)
        assert_raises(ValueError, _CheckInputs(x, y))

    def test_error_lowsamples(self):
        # raises error if samples are low (< 3)
        x = np.arange(3)
        y = np.arange(3)
        assert_raises(ValueError, _CheckInputs(x, y))

    def test_error_nans(self):
        # raises error if inputs contain NaNs
        x = np.arange(20, dtype=float)
        x[0] = np.nan
        assert_raises(ValueError, _CheckInputs(x, x))

        y = np.arange(20)
        assert_raises(ValueError, _CheckInputs(x, y))

    @pytest.mark.parametrize(
        "reps", [-1, "1"]  # reps is negative  # reps is not integer
    )
    def test_error_reps(self, reps):
        # raises error if reps is negative
        x = np.arange(20)
        assert_raises(ValueError, _CheckInputs(x, x, reps=reps))


class TestHelper:
    def test_simmat(self):
        # raises error if x or y is not a ndarray
        clf = RandomForestRegressor()
        x = np.arange(20).reshape(-1, 1)
        y = np.arange(5, 25)
        clf.fit(x, y)
        kernx = sim_matrix(clf, x)
        _check_kernmat(kernx, kernx)



import numpy as np
import pytest
from numpy.testing import assert_equal, assert_raises

from .. import SIMULATIONS, gaussian_3samp, ksamp_sim, rot_ksamp


class TestKSampleSimShape:
    @pytest.mark.parametrize("n", [100, 1000])
    @pytest.mark.parametrize("p", [1, 5])
    @pytest.mark.parametrize(
        "indep_sim",
        SIMULATIONS.keys(),
    )
    @pytest.mark.parametrize("k, degree", [(2, 90), (3, [90, 90])])
    def test_shapes(self, indep_sim, k, n, p, degree):
        np.random.seed(123456789)
        sims = rot_ksamp(indep_sim, n, p, k=k, degree=degree)
        sims1 = rot_ksamp(indep_sim, n, p, k=k, degree=degree, pow_type="dim")
        sims2 = ksamp_sim("rot_ksamp", n, sim=indep_sim, p=p, k=k, degree=degree)

        if indep_sim in [
            "joint_normal",
            "logarithmic",
            "sin_four_pi",
            "sin_sixteen_pi",
            "two_parabolas",
            "square",
            "diamond",
            "circle",
            "ellipse",
            "multiplicative_noise",
            "multimodal_independence",
        ]:
            [assert_equal(sim.shape[1], p * 2) for sim in sims]
            [assert_equal(sim.shape[1], p * 2) for sim in sims1]
            [assert_equal(sim.shape[1], p * 2) for sim in sims2]
        else:
            [assert_equal(sim.shape[1], p + 1) for sim in sims]
            [assert_equal(sim.shape[1], p + 1) for sim in sims1]
            [assert_equal(sim.shape[1], p + 1) for sim in sims2]


class TestGaussianSimShape:
    @pytest.mark.parametrize("n", [100, 1000])
    @pytest.mark.parametrize(
        "case",
        [1, 2, 3, 4, 5],
    )
    def test_shapes(self, n, case):
        np.random.seed(123456789)
        sims = gaussian_3samp(n, case=case)

        [assert_equal(sim.shape[0], n) for sim in sims]


class TestKSampleSimErrorWarn:
    """Tests errors and warnings."""

    def test_wrong_powtype(self):
        assert_raises(ValueError, rot_ksamp, sim="linear", n=100, p=1, pow_type="abcd")

    def test_wrong_k(self):
        assert_raises(
            ValueError, rot_ksamp, sim="linear", k=3, n=100, p=1, pow_type="abcd"
        )
        assert_raises(
            ValueError,
            rot_ksamp,
            sim="linear",
            degree=[90, 90],
            n=100,
            p=1,
            pow_type="abcd",
        )



import numpy as np
from scipy.signal import correlate
from scipy.stats import chi2

from ._utils import _CheckInputs
from .base import TimeSeriesTest, TimeSeriesTestOutput


class LjungBox(TimeSeriesTest):
    r"""
    Ljung-Box for Cross Correlation (CorrX) test statistic and p-value.

    Parameters
    ----------
    max_lag : int, default: 0
        The maximum number of lags in the past to check dependence between ``x`` and the
        shifted ``y``. If ``None``, then ``max_lag=np.ceil(np.log(n))``. Also the
        ``M`` hyperparmeter below.

    Notes
    -----
    The statistic can be derived as follows
    :footcite:p:`ljung1978measure`:

    Let :math:`x` and :math:`y` be :math:`(n, 1)` and :math:`(n, 1)` series
    respectively, which each contain :math:`y` observations of the series
    :math:`(X_t)` and :math:`(Y_t)`. Similarly, let :math:`x[j:n]` be the
    :math:`(n-j, p)` last :math:`n-j` observations of :math:`x`. Let :math:`y[0:(n-j)]`
    be the :math:`(n-j, p)` first :math:`n-j` observations of :math:`y`. Let :math:`M`
    be the maximum lag hyperparameter. The cross distance correlation is,

    .. math::

        \mathrm{Ljung-Box}_n (x, y) =  n(n+2)\sum_{j=1}^M \frac{
            \rho^2(x[j:n], y[0:(n-j)])}{n-j}

    where :math:`\rho` is the Pearson correlation coefficient.
    The p-value returned is calculated either via chi-squared distribution or
    using a permutation test.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, max_lag=0):
        self.is_distance = False
        TimeSeriesTest.__init__(self, compute_distance=None, max_lag=max_lag)

    def statistic(self, x, y):
        r"""
        Helper function that calculates the Ljung-Box cross correlation test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, 1)`` and ``(n, 1)`` where
            `n` is the number of samples.

        Returns
        -------
        stat : float
            The computed Ljung-Box statistic.
        opt_lag : int
            The computed optimal lag.
        """
        n = x.shape[0]

        ccf = correlate(x - x.mean(), y - y.mean())[n - 1 :].ravel()
        ccf /= np.arange(n, 0, -1)
        ccf /= np.std(x) * np.std(y)
        stat = (
            ccf[1 : self.max_lag + 1] ** 2
            / (n - np.arange(1, self.max_lag + 1))
            * (n * (n + 2))
        )

        self.stat = np.sum(stat)
        self.opt_lag = np.argmax(stat) + 1

        return self.stat, self.opt_lag

    def test(self, x, y, reps=1000, workers=1, auto=True, random_state=None):
        check_input = _CheckInputs(
            x,
            y,
            max_lag=self.max_lag,
        )
        x, y, self.max_lag = check_input()

        if self.max_lag <= 0:
            raise ValueError("max_lag must be greater than 0")

        # calculate observed test statistic
        stat, opt_lag = self.statistic(x, y)

        if auto and x.shape[0] > 20:
            pvalue = chi2.sf(stat, self.max_lag)
            self.null_dist = None
        else:
            stat, pvalue, stat_list = super(LjungBox, self).test(
                x=x,
                y=y,
                reps=reps,
                workers=workers,
                random_state=random_state,
                is_distsim=False,
            )

        self.pvalue = pvalue
        test_dict = {"opt_lag": opt_lag}

        return TimeSeriesTestOutput(stat, pvalue, test_dict)



import numpy as np
import pytest
from numpy.testing import assert_equal, assert_raises

from .. import ts_sim


class TestTSSimShape:
    @pytest.mark.parametrize("n", [100, 1000])
    @pytest.mark.parametrize(
        "sim",
        ["indep_ar", "cross_corr_ar", "nonlinear_process", "extinct_gaussian_process"],
    )
    def test_shapes(self, n, sim):
        np.random.seed(123456789)
        x, y = ts_sim(sim, n)
        assert_equal(x.shape, y.shape)


class TestTSSimErrorWarn:
    """Tests errors and warnings."""

    def test_np_inctype(self):
        assert_raises(ValueError, ts_sim, "indep_ar", n="a")

    def test_low_n(self):
        assert_raises(ValueError, ts_sim, "indep_ar", n=1)

    def test_extra_inctype(self):
        assert_raises(ValueError, ts_sim, "indep_ar", n=10, phi="a")

    def test_wrong_sim(self):
        assert_raises(ValueError, ts_sim, "abcd", n=10, phi=1)



import os
import sys

from setuptools import find_packages, setup
from setuptools.command.install import install

PACKAGE_NAME = "hyppo"
DESCRIPTION = "A comprehensive independence testing package"
LONG_DESCRIPTION = """
hyppo (**HYP**othesis Testing in **P**yth**O**n, pronounced "Hippo") is an open-source
software package for multivariate hypothesis testing. We decided to develop hyppo for
the following reasons:

* With the increase in the amount of data in many fields, hypothesis testing for high
dimensional and nonlinear data is important
* Libraries in R exist, but their interfaces are inconsistent and most are not available
in Python

hyppo intends to be a comprehensive multivariate hypothesis testing package and runs on
all major versions of operating systems. It also includes novel tests not found in other
packages. It is quick to install and free of charge. If you need to use multivariate
hypothesis testing, be sure to give hyppo a try!"
"""
AUTHOR = ("Sambit Panda",)
AUTHOR_EMAIL = "spanda3@jhu.edu"
URL = "https://github.com/neurodata/hyppo"
MINIMUM_PYTHON_VERSION = 3, 6  # Minimum of Python 3.6
REQUIRED_PACKAGES = [
    "numpy>=1.17",
    "scipy>=1.4.0",
    "numba>=0.46",
    "scikit-learn>=0.19.1",
    "autograd>=1.3",
]

# Find mgc version.
PROJECT_PATH = os.path.dirname(os.path.abspath(__file__))
for line in open(os.path.join(PROJECT_PATH, "hyppo", "__init__.py")):
    if line.startswith("__version__ = "):
        VERSION = line.strip().split()[2][1:-1]


def check_python_version():
    """Exit when the Python version is too low."""
    if sys.version_info < MINIMUM_PYTHON_VERSION:
        sys.exit("Python {}.{}+ is required.".format(*MINIMUM_PYTHON_VERSION))


check_python_version()


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""

    description = "verify that the git tag matches our version"

    def run(self):
        tag = os.getenv("CIRCLE_TAG")
        version = "v{}".format(VERSION)

        if tag != version:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, version
            )
            sys.exit(info)


setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    install_requires=REQUIRED_PACKAGES,
    url=URL,
    license="PolyForm Noncommercial License 1.0.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Mathematics",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    packages=find_packages(),
    include_package_data=True,
    test_suite="tests",
    cmdclass={
        "verify": VerifyVersionCommand,
    },
)



from typing import NamedTuple

import numpy as np
from scipy._lib._util import MapWrapper
from ._utils import _CheckInputs
from .base import DiscriminabilityTest
from sklearn.utils import check_random_state


class DiscrimOneSampleTestOutput(NamedTuple):
    stat: float
    pvalue: float
    null_dist: float


class DiscrimOneSample(DiscriminabilityTest):
    r"""

    One Sample Discriminability test statistic and p-value.

    Discriminability index is a measure of whether a data acquisition and
    preprocessing pipeline is more discriminable among different subjects.
    The key insight is that each repeated mesurements of the same item should
    be the more similar to one another than measurements between different
    items. The one sample test measures whether the discriminability
    for a dataset differs from random chance. More details are in `[1]`_.

    With :math:`D_x` as the sample discriminability of :math:`x`,
    one sample test performs the following test,

    .. math::

        H_0: D_x &= D_0 \\
        H_A: D_x &> D_0

    where :math:`D_0` is the discriminability that would be observed by random chance.

    .. _[1]: https://www.biorxiv.org/content/10.1101/802629v1

    Parameters
    ----------
    is_dist : bool, default: False
        Whether inputs are distance matrices.
    remove_isolates : bool, default: True
        Whether to remove the measurements with a single instance or not.
    """

    def __init__(self, is_dist=False, remove_isolates=True):
        # set is_distance to true if compute_distance is None
        self.is_distance = is_dist
        self.remove_isolates = remove_isolates
        DiscriminabilityTest.__init__(self)

    def statistic(self, x, y):
        """
        Helper function that calculates the discriminability test statistics.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.

        Returns
        -------
        stat : float
            The computed two sample discriminability statistic.
        """
        stat = super(DiscrimOneSample, self).statistic(x, y)

        return stat

    def test(self, x, y, reps=1000, workers=1, random_state=None):
        r"""
        Calculates the test statistic and p-value for Discriminability one sample test.

        Parameters
        ----------
        x : ndarray of float
            Input data matrices. ``x`` must have shape ``(n, p)`` where `n` is the
            number of
            samples and `p` are the number of dimensions. Alternatively, ``x`` can be
            distance matrices, where the shape must be ``(n, n)``, and ``is_dist`` must
            set to ``True`` in this case.
        y : ndarray of float
            A vector containing the sample ids for our `n` samples.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.

        Returns
        -------
        stat : float
            The computed discriminability statistic.
        pvalue : float
            The computed one sample test p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.discrim import DiscrimOneSample
        >>> x = np.concatenate([np.zeros((50, 2)), np.ones((50, 2))], axis=0)
        >>> y = np.concatenate([np.zeros(50), np.ones(50)], axis=0)
        >>> '%.1f, %.2f' % DiscrimOneSample().test(x, y) # doctest: +SKIP
        '1.0, 0.00'
        """

        check_input = _CheckInputs(
            [x],
            y,
            reps=reps,
            is_dist=self.is_distance,
            remove_isolates=self.remove_isolates,
        )
        x, y = check_input()

        self.x = np.asarray(x[0])
        self.y = y

        stat = self.statistic(self.x, self.y)
        self.stat = stat

        random_state = np.random.randint(np.iinfo(np.int32).max, size=reps)
        # use all cores to create function that parallelizes over number of reps
        mapwrapper = MapWrapper(workers)
        null_dist = np.array(list(mapwrapper(self._perm_stat, random_state)))
        self.null_dist = null_dist

        # calculate p-value and significant permutation map through list
        pvalue = (1 + (null_dist >= stat).sum()) / (1 + reps)

        self.pvalue_ = pvalue

        return DiscrimOneSampleTestOutput(stat, pvalue, null_dist)

    def _perm_stat(self, index, random_state=None):  # pragma: no cover
        r"""
        Helper function that is used to calculate parallel permuted test
        statistics.

        Parameters
        ----------
        index : int
            Iterator used for parallel statistic calculation

        Returns
        -------
        perm_stat : float
            Test statistic for each value in the null distribution.
        """
        rng = check_random_state(random_state)
        permy = rng.permutation(self.y)

        perm_stat = self.statistic(self.x, permy)

        return perm_stat



import numpy as np
from scipy.stats import gamma
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.gaussian_process.kernels import RBF

from .base import ConditionalIndependenceTest, ConditionalIndependenceTestOutput


class KCI(ConditionalIndependenceTest):
    r"""
    Kernel Conditional Independence Test Statistic and P-Value.

    This is a conditional indpendence test utilizing a radial basis
    function to calculate the kernels of two datasets. The trace
    of the normalized matrix product is then calculated to extract the test
    statistic. A Gaussian distribution is then utilized to calculate
    the p-value given the statistic and approximate mean and variance
    of the trace values of the independent kernel matrices.
    This test is consistent against similar tests.

    Notes
    -----
    The statistic is computed as follows :footcite:p:`10.5555/3020548.3020641`:

    Let :math:`x` be a combined sample of :math:`(n, p)` sample
    of random variables :math:`X` and let :math:`y` be a :math:`(n, 1)`
    labels of sample classes :math:`Y`. We can then generate
    :math:`K^x` and :math:`K^y` kernel matrices for each of the respective
    samples. Normalizing, multiplying, and taking the trace of these
    kernel matrices gives the resulting test statistic.
    The p-value and null distribution for the corrected statistic are calculated a
    gamma distribution approximation.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, **kwargs):

        ConditionalIndependenceTest.__init__(self, **kwargs)

    def compute_kern(self, x, y):

        T = len(y)

        x = np.array(x)
        y = np.array(y)
        x = x - np.mean(x)
        x = x / np.std(x)
        y = y - np.mean(y)
        y = y / np.std(y)

        if T < 200:
            width = 0.8
        elif T < 1200:
            width = 0.5
        else:
            width = 0.3

        theta = 1 / (width**2)

        Kx = 1.0 * RBF(theta).__call__(x, x)
        Ky = 1.0 * RBF(theta).__call__(y, y)

        return Kx, Ky

    def statistic(self, x, y):

        T = len(y)

        H = np.eye(T) - np.ones((T, T)) / T

        Kx, Ky = self.compute_kern(x, y)

        Kx = (H @ Kx) @ H
        Ky = (H @ Ky) @ H

        stat = np.trace(Kx @ Ky)

        return stat

    def test(self, x, y):
        r"""
        Calculates the Kernel Conditional Independence test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            columns. That is, the shapes must be ``(n, p)`` and ``(n, 1)`` where
            `n` is the dimension of samples and `p` is the number of
            dimensions.

        Returns
        -------
        stat : float
            The computed Kernel Conditional Independence statistic.
        pvalue : float
            The computed Kernel Conditional Independence p-value.

        Example
        --------
        >>> from hyppo.conditional import KCI
        >>> from hyppo.tools.indep_sim import linear
        >>> np.random.seed(123456789)
        >>> x, y = linear(100, 1)
        >>> stat, pvalue = KCI().test(x, y)
        >>> '%.1f, %.2f' % (stat, pvalue)
        '544.7, 0.00'
        """

        T = len(y)

        Kx, Ky = self.compute_kern(x, y)
        stat = self.statistic(x, y)

        mean_appr = (np.trace(Kx) * np.trace(Ky)) / T
        var_appr = (
            2 * np.trace(Kx @ Kx) * np.trace(Ky @ Ky) / T**2
        )
        k_appr = mean_appr**2 / var_appr
        theta_appr = var_appr / mean_appr
        pvalue = 1 - np.mean(gamma.cdf(stat, k_appr, theta_appr))

        self.stat = stat

        return ConditionalIndependenceTestOutput(stat, pvalue)



import numpy as np
import pytest
from numpy.testing import assert_almost_equal
from sklearn.metrics import pairwise_distances

from ...tools import linear, power
from .. import HHG
from ..hhg import hoeffdings


class TestFastHHGStat:
    @pytest.mark.parametrize("n, obs_stat", [(10, 1.0), (50, 1.0), (100, 1.0)])
    @pytest.mark.parametrize("obs_pvalue", [1 / 1000])
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = linear(n, 1)
        stat, pvalue = HHG().test(x, y, auto=True)

        assert_almost_equal(stat, obs_stat, decimal=2)
        assert_almost_equal(pvalue, obs_pvalue, decimal=2)

    def test_diststat(self):
        np.random.seed(123456789)
        x, y = linear(100, 1)
        zx = np.mean(x, axis=0).reshape(1, -1)
        zy = np.mean(y, axis=0).reshape(1, -1)
        distx = pairwise_distances(zx, x).reshape(-1, 1)
        disty = pairwise_distances(zy, y).reshape(-1, 1)
        test = HHG(compute_distance=None)
        test.auto = True
        stat = test.statistic(distx, disty)

        assert_almost_equal(stat, 1.0, decimal=2)

    def test_stat(self):
        np.random.seed(123456789)
        x, y = linear(100, 1)
        test = HHG()
        test.auto = True
        stat = test.statistic(x, y)

        assert_almost_equal(stat, 1.0, decimal=2)

    @pytest.mark.parametrize("n", [10, 200])
    def test_rep(self, n):
        x, y = linear(n, 1)
        stat, pvalue = HHG().test(x, y, auto=True, random_state=2)
        stat2, pvalue2 = HHG().test(x, y, auto=True, random_state=2)

        assert stat == stat2
        assert pvalue == pvalue2



from abc import ABC, abstractmethod

import numpy as np


class DiscriminabilityTest(ABC):
    r"""
    A base class for a discriminability test.
    """

    def __init__(self):
        self.pvalue_ = None
        super().__init__()

    @abstractmethod
    def statistic(self, x, y):
        r"""
        Calulates the independence test statistic.

        Parameters
        ----------
        x, y : ndarray of float
            Input data matrices.
        """

        rdfs = _discr_rdf(x, y)
        stat = np.nanmean(rdfs)

        return stat

    def _perm_stat(self, index):
        r"""
        Helper function that is used to calculate parallel permuted test
        statistics.

        Parameters
        ----------
        index : int
            Iterator used for parallel statistic calculation

        Returns
        -------
        perm_stat : float
            Test statistic for each value in the null distribution.
        """

    @abstractmethod
    def test(self, *args, **kwargs):
        r"""
        Calculates the Discriminability test statistic and p-value.
        """


def _discr_rdf(dissimilarities, labels):
    """
    Calculates the distributions of the test statistics.

    Parameters
    ----------
    dissimilarities : ndarray of float
        Input matrix containing the disimilarities.
    labels : ndarray of float
        The label matrix corresponding to each dissimilarity.

    Returns
    -------
    [type]
        [description]
    """
    # calculates test statistics distribution
    rdfs = []

    for i, label in enumerate(labels):
        di = dissimilarities[i]

        # All other samples except its own label
        idx = labels == label
        Dij = di[~idx]

        # All samples except itself
        idx[i] = False
        Dii = di[idx]

        rdf = [1 - ((Dij < d).sum() + 0.5 * (Dij == d).sum()) / Dij.size for d in Dii]
        rdfs.append(rdf)

    out = np.full((len(rdfs), max(map(len, rdfs))), np.nan)
    for i, rdf in enumerate(rdfs):
        out[i, : len(rdf)] = rdf

    return out
